from pickle import dump, load

from ..core import PPQ_CONFIG
from ..IR import BaseGraph, GraphExporter
from ..IR.base.graph import GraphBuilder


class NativeExporter(GraphExporter):
    def __init__(self) -> None:
        super().__init__()

    def export(self, file_path: str, graph: BaseGraph, config_path: str = None, dump_value: bool = True):
        def dump_elements_to_file(file, elements: list):
            for element in elements:
                dump(element, file)

        with open(file_path, "wb") as file:
            dump_elements_to_file(
                file,
                elements=[
                    "PPQ GRAPH DEFINITION",  # PPQ Signature.
                    PPQ_CONFIG.VERSION,  # PPQ Signature.
                    graph,
                ],
            )


class NativeImporter(GraphBuilder):
    def __init__(self) -> None:
        super().__init__()

    def build(self, file_path: str, **kwargs) -> BaseGraph:
        def load_elements_from_file(file, num_of_elements: int) -> list:
            try:
                return [load(file) for _ in range(num_of_elements)]
            except EOFError as e:
                raise Exception("File format parsing error. Unexpected EOF found.")

        with open(file_path, "rb") as file:
            signature, version, graph = load_elements_from_file(file, 3)
            if signature != "PPQ GRAPH DEFINITION":
                raise Exception("File format parsing error. Graph Signature has been damaged.")
            if str(version) > PPQ_CONFIG.VERSION:
                print(
                    f"\033[31mWarning: Dump file is created by PPQ({str(version)}), "
                    f"however you are using PPQ({PPQ_CONFIG.VERSION}).\033[0m"
                )

        assert isinstance(graph, BaseGraph), "File format parsing error. Graph Definition has been damaged."
        try:
            for op in graph.operations.values():
                input_copy, _ = op.inputs.copy(), op.inputs.clear()
                for name in input_copy:
                    op.inputs.append(graph.variables[name])
                output_copy, _ = op.outputs.copy(), op.outputs.clear()
                for name in output_copy:
                    op.outputs.append(graph.variables[name])

            for var in graph.variables.values():
                dest_copy, _ = var.dest_ops.copy(), var.dest_ops.clear()
                for name in dest_copy:
                    var.dest_ops.append(graph.operations[name])
                if var.source_op is not None:
                    var.source_op = graph.operations[var.source_op]

            graph._graph_inputs = {name: graph.variables[name] for name in graph._graph_inputs}
            graph._graph_outputs = {name: graph.variables[name] for name in graph._graph_outputs}
        except Exception as e:
            raise Exception("File format parsing error. Graph Definition has been damaged.")
        return graph



from abc import ABCMeta, abstractmethod
from typing import Any

from ...core import QuantizationStates, TensorQuantizationConfig, ppq_debug_function
from ...IR import Variable


class BaseTensorObserver(metaclass=ABCMeta):
    def __init__(self, watch_on: Variable, quant_cfg: TensorQuantizationConfig):
        self._watch_on = watch_on
        self._quant_cfg = quant_cfg

    @abstractmethod
    def observe(self, value: Any):
        raise NotImplementedError("Implement this function first.")

    @abstractmethod
    def render_quantization_config(self):
        raise NotImplementedError("Implement this function first.")

    def __str__(self) -> str:
        return (
            "PPQ Tensor Observer ("
            + self.__class__.__name__
            + ") mount on variable "
            + self._watch_on.name
            + " observing algorithm: "
            + self._quant_cfg.observer_algorithm
        )

    @ppq_debug_function
    def report(self) -> str:
        if self._quant_cfg.state == QuantizationStates.ACTIVATED:
            return (
                f"Observer on Variable {self._watch_on.name}, "
                f"computed scale: {self._quant_cfg.scale}, computed offset: {self._quant_cfg.offset}\n"
            )



import torch
from torch.autograd import Function
from ...core import PPQ_CONFIG, QuantizationProperty, QuantizationStates, RoundingPolicy, TensorQuantizationConfig
from ...utils.round import ppq_tensor_round


class TensorwiseLinearQuantImpl(Function):
    """Torch Tensorwise quantize is designed to quantize a torch Tensor
    with a given configuration. All quantization within PPQ will invoke
    this function to quantize its value. Any modification of this function
    will greatly affects system behaviour.

    This is a torch implementation of quantization itself.
    Notice that if ppq.config.USING_CUDA_KERNAL = True,
        then all quantization will use ffi.CUDA instead.

    Notice this function will always clone your tensor value first.
    This function never quantize your tensor value inplace.
    """

    @staticmethod
    def forward(
        ctx,
        tensor: torch.Tensor,
        scales: torch.Tensor,
        offsets: torch.Tensor,
        quant_min: int,
        quant_max: int,
        rounding: RoundingPolicy,
    ) -> torch.Tensor:
        scales, offsets = scales.to(tensor.device), offsets.to(tensor.device)

        if not PPQ_CONFIG.USING_CUDA_KERNEL or not tensor.is_cuda:
            # quantization function, pytorch implmentation
            tensor = ppq_tensor_round((tensor / scales), rounding) + offsets
            tensor = torch.clamp(tensor, quant_min, quant_max)
            tensor = (tensor - offsets) * scales
            return tensor

        else:
            raise NotImplementedError()

    @staticmethod
    def backward(ctx, dy: torch.Tensor):
        return dy, None, None, None, None, None, None, None, None


class ChannelwiseLinearQuantImpl(Function):
    """Torch Channelwise quantize is designed to quantize a torch Tensor
    with a given configuration. All quantization within PPQ will invoke
    this function to quantize its value. Any modification of this function
    will greatly affects system behaviour.

    This is a torch implementation of quantization itself.
    Notice that if ppq.config.USING_CUDA_KERNAL = True,
        then all quantization will bypass this function by using ffi.CUDA instead.

    Notice this function will always clone your tensor value first.
    This function never quantize your tensor value inplace.
    """

    @staticmethod
    def forward(
        ctx,
        tensor: torch.Tensor,
        scales: torch.Tensor,
        offsets: torch.Tensor,
        channel_axis: int,
        quant_min: int,
        quant_max: int,
        rounding: RoundingPolicy,
    ) -> torch.Tensor:
        scales, offsets = scales.to(tensor.device), offsets.to(tensor.device)

        if not PPQ_CONFIG.USING_CUDA_KERNEL or not tensor.is_cuda:
            # generate a shape that likes [1, 1, -1, 1], the only -1 is at channel axe.
            shape = [1 if axis != channel_axis else -1 for axis in range(tensor.ndim)]
            scale, offset = scales.view(shape), offsets.view(shape)

            tensor = ppq_tensor_round((tensor / scale), rounding) + offset
            tensor = torch.clamp(tensor, quant_min, quant_max)
            tensor = (tensor - offset) * scale
            return tensor
        else:
            raise NotImplementedError()

    @staticmethod
    def backward(ctx, dy: torch.Tensor):
        return dy, None, None, None, None, None, None, None, None, None


class TensorwiseDynamicLinearQuantImpl(Function):
    """Torch Tensorwise quantize is designed to quantize a torch Tensor
    with a given configuration. All quantization within PPQ will invoke
    this function to quantize its value. Any modification of this function
    will greatly affects system behaviour.

    This is a torch implementation of quantization itself.
    Notice that if ppq.config.USING_CUDA_KERNAL = True,
        then all quantization will use ffi.CUDA instead.

    Notice this function will always clone your tensor value first.
    This function never quantize your tensor value inplace.
    """

    @staticmethod
    def forward(ctx, tensor: torch.Tensor, config: TensorQuantizationConfig) -> torch.Tensor:
        from ..observer.range import minmax_to_scale_offset

        # solve scale and offset at first.
        scales, offsets = minmax_to_scale_offset(tensor.min().item(), tensor.max().item(), config=config)
        print(scales, offsets)
        # quantization function, pytorch implmentation
        tensor = ppq_tensor_round((tensor / scales), config.rounding) + offsets
        tensor = torch.clamp(tensor, config.quant_min, config.quant_max)
        tensor = (tensor - offsets) * scales
        return tensor

    @staticmethod
    def backward(ctx, dy: torch.Tensor):
        return dy, None


class ChannelwiseDynamicLinearQuantImpl(Function):
    """Torch Channelwise quantize is designed to quantize a torch Tensor
    with a given configuration. All quantization within PPQ will invoke
    this function to quantize its value. Any modification of this function
    will greatly affects system behaviour.

    This is a torch implementation of quantization itself.
    Notice that if ppq.config.USING_CUDA_KERNAL = True,
        then all quantization will bypass this function by using ffi.CUDA instead.

    Notice this function will always clone your tensor value first.
    This function never quantize your tensor value inplace.
    """

    @staticmethod
    def forward(ctx, tensor: torch.Tensor, config: TensorQuantizationConfig) -> torch.Tensor:
        from ..observer.range import minmax_to_scale_offset

        channelwise_view = tensor.transpose(dim0=0, dim1=config.channel_axis).unsqueeze(-1)
        channelwise_view = torch.flatten(channelwise_view, start_dim=1)

        scales, offsets = [], []
        for _min, _max in zip(channelwise_view.min(dim=1)[0].tolist(), channelwise_view.max(dim=1)[0].tolist()):
            s, o = minmax_to_scale_offset(_min, _max, config)
            scales.append(s)
            offsets.append(o)

        scales = torch.tensor(scales, dtype=torch.float32, device=tensor.device)
        offsets = torch.tensor(offsets, dtype=torch.float32, device=tensor.device)

        # generate a shape that likes [1, 1, -1, 1], the only -1 is at channel axe.
        shape = [1 if axis != config.channel_axis else -1 for axis in range(tensor.ndim)]
        scales, offsets = scales.view(shape), offsets.view(shape)

        tensor = ppq_tensor_round((tensor / scales), config.rounding) + offsets
        tensor = torch.clamp(tensor, config.quant_min, config.quant_max)
        tensor = (tensor - offsets) * scales
        return tensor

    @staticmethod
    def backward(ctx, dy: torch.Tensor):
        return dy, None


def PPQDyamicLinearQuantFunction(tensor: torch.Tensor, config: TensorQuantizationConfig) -> torch.Tensor:
    """
    Dynamic Linear Quantization Function(PPQ 动态量化函数).

    When calling this method, we firstly solve a scale & offset setting by min-max observer.

    Then we applys ordinary Linear Quantization Function with solved setting.

    If there is a pre-defined scale & offset within given config, they will be dropped without warning.

    动态量化函数将在执行量化之前统计出 tensor 的 min - max, 而后计算出 scale & offset 并完成量化

    此时 TQC 中的 scale 与 offset 将被忽略
    """
    if not QuantizationStates.is_activated(config.state):
        return tensor
    if not config.policy.has_property(QuantizationProperty.LINEAR):
        raise ValueError("Critical Quantization Error! Non-linear config detected.")
    if not config.policy.has_property(QuantizationProperty.DYNAMIC):
        raise ValueError("Quantization Policy Do Not Have Dynamic Flag!")

    if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
        return ChannelwiseDynamicLinearQuantImpl.apply(tensor, config)
    elif config.policy.has_property(QuantizationProperty.PER_TENSOR):
        return TensorwiseDynamicLinearQuantImpl.apply(tensor, config)


def PPQLinearQuantFunction(tensor: torch.Tensor, config: TensorQuantizationConfig) -> torch.Tensor:
    """PPQ 核心量化函数，没啥好说的了吧，这个玩意既做 quant 也做 dequant"""
    if not QuantizationStates.is_activated(config.state):
        return tensor
    if not config.policy.has_property(QuantizationProperty.LINEAR):
        raise ValueError("Critical Quantization Error! Non-linear config detected.")
    if config.policy.has_property(QuantizationProperty.DYNAMIC):
        raise ValueError("Unexpected Dynamic Flag in Quantization Policy. Use PPQDyamicQuantFunction Instead.")

    if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
        return ChannelwiseLinearQuantImpl.apply(
            tensor,
            config.scale,
            config.offset,
            config.channel_axis,
            config.quant_min,
            config.quant_max,
            config.rounding,
        )
    elif config.policy.has_property(QuantizationProperty.PER_TENSOR):
        return TensorwiseLinearQuantImpl.apply(
            tensor, config.scale, config.offset, config.quant_min, config.quant_max, config.rounding
        )


def PPQLinearQuant_toInt(tensor: torch.Tensor, config: TensorQuantizationConfig) -> torch.Tensor:
    """PPQ 核心量化函数，没啥好说的了吧，这个玩意只做 quant 不做 dequant"""
    if not config.policy.has_property(QuantizationProperty.LINEAR):
        raise ValueError("Critical Quantization Error! Non-linear config detected.")
    if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
        shape = [1 if axis != config.channel_axis else -1 for axis in range(tensor.ndim)]
        scale, offset = config.scale.view(shape), config.offset.view(shape)
        tensor = ppq_tensor_round((tensor / scale), config.rounding) + offset
        tensor = torch.clamp(tensor, config.quant_min, config.quant_max)
    elif config.policy.has_property(QuantizationProperty.PER_TENSOR):
        tensor = ppq_tensor_round((tensor / config.scale), config.rounding) + config.offset
        tensor = torch.clamp(tensor, config.quant_min, config.quant_max)

    if config.num_of_bits == 8:
        if config.policy.has_property(QuantizationProperty.ASYMMETRICAL) and config.quant_min == 0:
            return tensor.type(dtype=torch.uint8)
        else:
            return tensor.type(dtype=torch.int8)
    elif config.num_of_bits > 8 and config.num_of_bits <= 16:
        return tensor.type(dtype=torch.int16)
    elif config.num_of_bits > 16:
        return tensor.type(dtype=torch.int32)
    else:
        raise Exception("Do not konw how to convert value into int. num of bits is unexpected.")



from .baking import ParameterBakingPass
from .base import QuantizationOptimizationPass, QuantizationOptimizationPipeline
from .equalization import ActivationEqualizationPass, ChannelwiseSplitPass, LayerwiseEqualizationPass
from .morph import (
    GRUSplitPass,
    HorizontalLayerSplitPass,
    MetaxGemmSplitPass,
    NCNNFormatGemmPass,
    NXPResizeModeChangePass,
)
from .parameters import ParameterQuantizePass, PassiveParameterQuantizePass
from .refine import (
    MishFusionPass,
    NxpInputRoundingRefinePass,
    NxpQuantizeFusionPass,
    QuantAlignmentPass,
    QuantizeFusionPass,
    QuantizeSimplifyPass,
    SwishFusionPass,
)
from .ssd import SSDEqualizationPass
from .training import BiasCorrectionPass, LearnedStepSizePass, RoundTuningPass



"""Lightweight unit tests for torch executor operator execution."""

import sys
import os
import unittest

import torch

sys.path.insert(0, os.path.join(os.path.dirname(__file__), ".."))

from xslim.ppq_decorator.ppq.core import DataType, TargetPlatform
from xslim.ppq_decorator.ppq.IR import BaseGraph, Operation, Variable
from xslim.ppq_decorator.ppq.IR.base.opdef import Opset
from xslim.ppq_decorator.ppq.executor.op import TorchBackendContext, DEFAULT_BACKEND_TABLE
from xslim.ppq_decorator.ppq.executor.torch import TorchExecutor


def make_op(name, op_type, attributes=None, num_inputs=1, num_outputs=1):
    """Helper to create a minimal Operation with linked Variables."""
    attributes = attributes or {}
    inputs = [Variable(name=f"{name}_in_{i}") for i in range(num_inputs)]
    outputs = [Variable(name=f"{name}_out_{i}") for i in range(num_outputs)]
    op = Operation(
        name=name,
        op_type=op_type,
        attributes=attributes,
        platform=TargetPlatform.UNSPECIFIED,
        inputs=inputs,
        outputs=outputs,
    )
    for v in inputs:
        v._dest_ops.append(op)
    for v in outputs:
        v._source_op = op
    return op


CTX = TorchBackendContext(executing_device="cpu")


class TestArithmeticOps(unittest.TestCase):
    """Test arithmetic operator forward functions."""

    def test_add(self):
        op = make_op("add", "Add", num_inputs=2)
        a = torch.tensor([1.0, 2.0, 3.0])
        b = torch.tensor([4.0, 5.0, 6.0])
        result = DEFAULT_BACKEND_TABLE["Add"](op, [a, b], CTX)
        torch.testing.assert_close(result, a + b)

    def test_add_broadcast(self):
        op = make_op("add_bc", "Add", num_inputs=2)
        a = torch.randn(2, 3)
        b = torch.randn(3)
        result = DEFAULT_BACKEND_TABLE["Add"](op, [a, b], CTX)
        torch.testing.assert_close(result, a + b)

    def test_mul(self):
        op = make_op("mul", "Mul", num_inputs=2)
        a = torch.tensor([2.0, 3.0])
        b = torch.tensor([4.0, 5.0])
        result = DEFAULT_BACKEND_TABLE["Mul"](op, [a, b], CTX)
        torch.testing.assert_close(result, a * b)

    def test_sub(self):
        op = make_op("sub", "Sub", num_inputs=2)
        a = torch.tensor([5.0, 3.0])
        b = torch.tensor([1.0, 2.0])
        result = DEFAULT_BACKEND_TABLE["Sub"](op, [a, b], CTX)
        torch.testing.assert_close(result, (a - b).float())

    def test_div(self):
        op = make_op("div", "Div", num_inputs=2)
        a = torch.tensor([6.0, 8.0])
        b = torch.tensor([2.0, 4.0])
        result = DEFAULT_BACKEND_TABLE["Div"](op, [a, b], CTX)
        torch.testing.assert_close(result, a / b)


class TestActivationOps(unittest.TestCase):
    """Test activation operator forward functions."""

    def test_relu(self):
        op = make_op("relu", "Relu")
        x = torch.tensor([-1.0, 0.0, 1.0, 2.0])
        result = DEFAULT_BACKEND_TABLE["Relu"](op, [x], CTX)
        torch.testing.assert_close(result, torch.relu(x))

    def test_sigmoid(self):
        op = make_op("sigmoid", "Sigmoid")
        x = torch.tensor([-2.0, 0.0, 2.0])
        result = DEFAULT_BACKEND_TABLE["Sigmoid"](op, [x], CTX)
        torch.testing.assert_close(result, torch.sigmoid(x))

    def test_exp(self):
        op = make_op("exp", "Exp")
        x = torch.tensor([0.0, 1.0, 2.0])
        result = DEFAULT_BACKEND_TABLE["Exp"](op, [x], CTX)
        torch.testing.assert_close(result, torch.exp(x))

    def test_tanh(self):
        op = make_op("tanh", "Tanh")
        x = torch.tensor([-1.0, 0.0, 1.0])
        result = DEFAULT_BACKEND_TABLE["Tanh"](op, [x], CTX)
        torch.testing.assert_close(result, torch.tanh(x))

    def test_softmax(self):
        op = make_op("softmax", "Softmax", attributes={"axis": -1})
        op._opset = Opset(version=13)
        x = torch.tensor([[1.0, 2.0, 3.0]])
        result = DEFAULT_BACKEND_TABLE["Softmax"](op, [x], CTX)
        torch.testing.assert_close(result, torch.softmax(x, dim=-1))

    def test_leaky_relu(self):
        op = make_op("leaky_relu", "LeakyRelu", attributes={"alpha": 0.01})
        x = torch.tensor([-2.0, -1.0, 0.0, 1.0])
        result = DEFAULT_BACKEND_TABLE["LeakyRelu"](op, [x], CTX)
        torch.testing.assert_close(result, torch.nn.functional.leaky_relu(x, 0.01))


class TestUnaryOps(unittest.TestCase):
    """Test unary operator forward functions."""

    def test_abs(self):
        op = make_op("abs", "Abs")
        x = torch.tensor([-3.0, -1.0, 0.0, 2.0])
        result = DEFAULT_BACKEND_TABLE["Abs"](op, [x], CTX)
        torch.testing.assert_close(result, x.abs())

    def test_sqrt(self):
        op = make_op("sqrt", "Sqrt")
        x = torch.tensor([1.0, 4.0, 9.0])
        result = DEFAULT_BACKEND_TABLE["Sqrt"](op, [x], CTX)
        torch.testing.assert_close(result, torch.sqrt(x))

    def test_neg(self):
        op = make_op("neg", "Neg")
        x = torch.tensor([1.0, -2.0, 3.0])
        result = DEFAULT_BACKEND_TABLE["Neg"](op, [x], CTX)
        torch.testing.assert_close(result, -x)

    def test_log(self):
        op = make_op("log", "Log")
        x = torch.tensor([1.0, 2.0, 3.0])
        result = DEFAULT_BACKEND_TABLE["Log"](op, [x], CTX)
        torch.testing.assert_close(result, torch.log(x))

    def test_floor(self):
        op = make_op("floor", "Floor")
        x = torch.tensor([1.5, 2.7, -0.3])
        result = DEFAULT_BACKEND_TABLE["Floor"](op, [x], CTX)
        torch.testing.assert_close(result, torch.floor(x))

    def test_reciprocal(self):
        op = make_op("reciprocal", "Reciprocal")
        x = torch.tensor([1.0, 2.0, 4.0])
        result = DEFAULT_BACKEND_TABLE["Reciprocal"](op, [x], CTX)
        torch.testing.assert_close(result, 1.0 / x)


class TestTensorManipulationOps(unittest.TestCase):
    """Test tensor manipulation operator forward functions."""

    def test_reshape(self):
        op = make_op("reshape", "Reshape", num_inputs=2)
        data = torch.randn(2, 3, 4)
        shape = torch.tensor([2, 12], dtype=torch.int64)
        result = DEFAULT_BACKEND_TABLE["Reshape"](op, [data, shape], CTX)
        self.assertEqual(result.shape, torch.Size([2, 12]))

    def test_transpose(self):
        op = make_op("transpose", "Transpose", attributes={"perm": [0, 2, 1]})
        x = torch.randn(2, 3, 4)
        result = DEFAULT_BACKEND_TABLE["Transpose"](op, [x], CTX)
        self.assertEqual(result.shape, torch.Size([2, 4, 3]))
        torch.testing.assert_close(result, x.permute(0, 2, 1))

    def test_concat(self):
        op = make_op("concat", "Concat", attributes={"axis": 0}, num_inputs=2)
        a = torch.randn(2, 3)
        b = torch.randn(3, 3)
        result = DEFAULT_BACKEND_TABLE["Concat"](op, [a, b], CTX)
        self.assertEqual(result.shape, torch.Size([5, 3]))

    def test_flatten(self):
        op = make_op("flatten", "Flatten", attributes={"axis": 1})
        x = torch.randn(2, 3, 4)
        result = DEFAULT_BACKEND_TABLE["Flatten"](op, [x], CTX)
        self.assertEqual(result.shape, torch.Size([2, 12]))

    def test_squeeze(self):
        op = make_op("squeeze", "Squeeze", num_inputs=2)
        op._opset = Opset(version=13)
        x = torch.randn(1, 3, 1, 4)
        axes = torch.tensor([0], dtype=torch.int64)
        result = DEFAULT_BACKEND_TABLE["Squeeze"](op, [x, axes], CTX)
        self.assertEqual(result.shape, torch.Size([3, 1, 4]))

    def test_unsqueeze(self):
        op = make_op("unsqueeze", "Unsqueeze", num_inputs=2)
        op._opset = Opset(version=13)
        x = torch.randn(3, 4)
        axes = torch.tensor([0], dtype=torch.int64)
        result = DEFAULT_BACKEND_TABLE["Unsqueeze"](op, [x, axes], CTX)
        self.assertEqual(result.shape, torch.Size([1, 3, 4]))


class TestMatrixOps(unittest.TestCase):
    """Test matrix operation forward functions."""

    def test_matmul_2d(self):
        op = make_op("matmul", "MatMul", num_inputs=2)
        a = torch.randn(3, 4)
        b = torch.randn(4, 5)
        result = DEFAULT_BACKEND_TABLE["MatMul"](op, [a, b], CTX)
        torch.testing.assert_close(result, torch.matmul(a, b))

    def test_matmul_batch(self):
        op = make_op("matmul_batch", "MatMul", num_inputs=2)
        a = torch.randn(2, 3, 4)
        b = torch.randn(2, 4, 5)
        result = DEFAULT_BACKEND_TABLE["MatMul"](op, [a, b], CTX)
        torch.testing.assert_close(result, torch.matmul(a, b))

    def test_gemm(self):
        op = make_op("gemm", "Gemm", attributes={"alpha": 1.0, "beta": 1.0, "transA": 0, "transB": 0}, num_inputs=3)
        a = torch.randn(3, 4)
        b = torch.randn(4, 5)
        c = torch.randn(5)
        result = DEFAULT_BACKEND_TABLE["Gemm"](op, [a, b, c], CTX)
        expected = torch.matmul(a, b) + c
        torch.testing.assert_close(result, expected)


class TestOtherOps(unittest.TestCase):
    """Test miscellaneous operator forward functions."""

    def test_clip(self):
        op = make_op("clip", "Clip", num_inputs=3)
        x = torch.tensor([-5.0, -1.0, 0.0, 3.0, 10.0])
        min_val = torch.tensor(-2.0)
        max_val = torch.tensor(5.0)
        result = DEFAULT_BACKEND_TABLE["Clip"](op, [x, min_val, max_val], CTX)
        torch.testing.assert_close(result, torch.clamp(x, -2.0, 5.0))

    def test_identity(self):
        op = make_op("identity", "Identity")
        x = torch.randn(3, 4)
        result = DEFAULT_BACKEND_TABLE["Identity"](op, [x], CTX)
        torch.testing.assert_close(result, x)

    def test_shape(self):
        op = make_op("shape", "Shape")
        x = torch.randn(2, 3, 4)
        result = DEFAULT_BACKEND_TABLE["Shape"](op, [x], CTX)
        expected = torch.tensor([2, 3, 4], dtype=torch.long)
        torch.testing.assert_close(result, expected)

    def test_where(self):
        op = make_op("where", "Where", num_inputs=3)
        cond = torch.tensor([True, False, True])
        a = torch.tensor([1.0, 2.0, 3.0])
        b = torch.tensor([4.0, 5.0, 6.0])
        result = DEFAULT_BACKEND_TABLE["Where"](op, [cond, a, b], CTX)
        torch.testing.assert_close(result, torch.where(cond, a, b))

    def test_cast_to_float(self):
        op = make_op("cast", "Cast", attributes={"to": DataType.FP32})
        x = torch.tensor([1, 2, 3], dtype=torch.int32)
        result = DEFAULT_BACKEND_TABLE["Cast"](op, [x], CTX)
        self.assertEqual(result.dtype, torch.float32)

    def test_constant_of_shape(self):
        op = make_op("const_shape", "ConstantOfShape", attributes={"value": torch.tensor([0.0])})
        shape = torch.tensor([2, 3], dtype=torch.int64)
        result = DEFAULT_BACKEND_TABLE["ConstantOfShape"](op, [shape], CTX)
        self.assertEqual(result.shape, torch.Size([2, 3]))


class TestReduceOps(unittest.TestCase):
    """Test reduce operator forward functions."""

    def test_reduce_mean(self):
        op = make_op("reduce_mean", "ReduceMean", attributes={"axes": [1], "keepdims": 1})
        op._opset = Opset(version=11)
        x = torch.randn(2, 3, 4)
        result = DEFAULT_BACKEND_TABLE["ReduceMean"](op, [x], CTX)
        expected = torch.mean(x, dim=1, keepdim=True)
        torch.testing.assert_close(result, expected)

    def test_reduce_sum(self):
        op = make_op("reduce_sum", "ReduceSum", attributes={"keepdims": 1}, num_inputs=2)
        op._opset = Opset(version=13)
        x = torch.randn(2, 3, 4)
        axes = torch.tensor([1], dtype=torch.int64)
        result = DEFAULT_BACKEND_TABLE["ReduceSum"](op, [x, axes], CTX)
        expected = torch.sum(x, dim=1, keepdim=True)
        torch.testing.assert_close(result, expected)


class TestConvOps(unittest.TestCase):
    """Test convolution operator forward functions."""

    def test_conv2d(self):
        op = make_op(
            "conv",
            "Conv",
            attributes={
                "kernel_shape": [3, 3],
                "strides": [1, 1],
                "pads": [1, 1, 1, 1],
                "dilations": [1, 1],
                "group": 1,
            },
            num_inputs=3,
        )
        x = torch.randn(1, 3, 8, 8)
        w = torch.randn(16, 3, 3, 3)
        b = torch.randn(16)
        result = DEFAULT_BACKEND_TABLE["Conv"](op, [x, w, b], CTX)
        self.assertEqual(result.shape, torch.Size([1, 16, 8, 8]))

    def test_conv2d_no_bias(self):
        op = make_op(
            "conv_nb",
            "Conv",
            attributes={
                "kernel_shape": [3, 3],
                "strides": [1, 1],
                "pads": [0, 0, 0, 0],
                "dilations": [1, 1],
                "group": 1,
            },
            num_inputs=2,
        )
        x = torch.randn(1, 3, 8, 8)
        w = torch.randn(16, 3, 3, 3)
        result = DEFAULT_BACKEND_TABLE["Conv"](op, [x, w], CTX)
        self.assertEqual(result.shape, torch.Size([1, 16, 6, 6]))


class TestTorchExecutorGraph(unittest.TestCase):
    """Test TorchExecutor with a minimal computation graph."""

    def _build_simple_graph(self):
        """Build: input -> Relu -> Add(with param) -> output."""
        graph = BaseGraph(name="test_graph")
        graph.set_extension_attrib("IS_DISPATCHED_GRAPH", True)

        # Variables
        input_var = Variable(name="input", shape=[1, 3], dtype=DataType.FP32)
        relu_out = Variable(name="relu_out", shape=[1, 3], dtype=DataType.FP32)
        param_var = Variable(name="param", value=torch.ones(1, 3), is_parameter=True, shape=[1, 3], dtype=DataType.FP32)
        output_var = Variable(name="output", shape=[1, 3], dtype=DataType.FP32)

        # Operations
        relu_op = Operation(name="relu_1", op_type="Relu", attributes={}, platform=TargetPlatform.UNSPECIFIED)
        add_op = Operation(name="add_1", op_type="Add", attributes={}, platform=TargetPlatform.UNSPECIFIED)

        # Link variables to operations
        input_var._dest_ops = [relu_op]
        relu_op._input_vars = [input_var]
        relu_op._output_vars = [relu_out]
        relu_out._source_op = relu_op

        relu_out._dest_ops = [add_op]
        param_var._dest_ops = [add_op]
        add_op._input_vars = [relu_out, param_var]
        add_op._output_vars = [output_var]
        output_var._source_op = add_op

        # Add variables first (those without dest_ops referencing ops not yet in graph)
        graph._variables["input"] = input_var
        graph._variables["param"] = param_var
        graph._variables["relu_out"] = relu_out
        graph._variables["output"] = output_var

        # Add operations
        graph._operations["relu_1"] = relu_op
        graph._operations["add_1"] = add_op

        # Set graph inputs and outputs
        graph._graph_inputs["input"] = input_var
        graph._graph_outputs["output"] = output_var

        return graph

    def test_executor_forward(self):
        graph = self._build_simple_graph()
        executor = TorchExecutor(graph=graph, device="cpu")

        x = torch.tensor([[-1.0, 2.0, -3.0]])
        results = executor.forward(inputs={"input": x}, output_names=["output"])

        expected = torch.relu(x) + torch.ones(1, 3)
        torch.testing.assert_close(results[0], expected)

    def test_executor_default_outputs(self):
        graph = self._build_simple_graph()
        executor = TorchExecutor(graph=graph, device="cpu")

        x = torch.tensor([[1.0, -1.0, 0.5]])
        results = executor.forward(inputs={"input": x})

        expected = torch.relu(x) + torch.ones(1, 3)
        torch.testing.assert_close(results[0], expected)

    def test_executor_list_input(self):
        graph = self._build_simple_graph()
        executor = TorchExecutor(graph=graph, device="cpu")

        x = torch.tensor([[0.5, -0.5, 1.0]])
        results = executor.forward(inputs=[x])

        expected = torch.relu(x) + torch.ones(1, 3)
        torch.testing.assert_close(results[0], expected)

    def test_executor_tensor_input(self):
        graph = self._build_simple_graph()
        executor = TorchExecutor(graph=graph, device="cpu")

        x = torch.tensor([[2.0, -2.0, 0.0]])
        results = executor.forward(inputs=x)

        expected = torch.relu(x) + torch.ones(1, 3)
        torch.testing.assert_close(results[0], expected)


class TestGridSampleOp(unittest.TestCase):
    """Test GridSample operator forward function."""

    def test_grid_sample_default(self):
        op = make_op("gs", "GridSample", num_inputs=2)
        x = torch.randn(1, 1, 4, 4)
        grid = torch.randn(1, 3, 3, 2).clamp(-1, 1)
        result = DEFAULT_BACKEND_TABLE["GridSample"](op, [x, grid], CTX)
        expected = torch.nn.functional.grid_sample(
            x, grid, mode="bilinear", padding_mode="zeros", align_corners=False
        )
        torch.testing.assert_close(result, expected)

    def test_grid_sample_nearest(self):
        op = make_op("gs_n", "GridSample", attributes={"mode": "nearest"}, num_inputs=2)
        x = torch.randn(1, 1, 4, 4)
        grid = torch.randn(1, 3, 3, 2).clamp(-1, 1)
        result = DEFAULT_BACKEND_TABLE["GridSample"](op, [x, grid], CTX)
        expected = torch.nn.functional.grid_sample(
            x, grid, mode="nearest", padding_mode="zeros", align_corners=False
        )
        torch.testing.assert_close(result, expected)

    def test_grid_sample_align_corners(self):
        op = make_op(
            "gs_ac", "GridSample",
            attributes={"align_corners": 1, "padding_mode": "border"},
            num_inputs=2,
        )
        x = torch.randn(1, 1, 4, 4)
        grid = torch.randn(1, 3, 3, 2).clamp(-1, 1)
        result = DEFAULT_BACKEND_TABLE["GridSample"](op, [x, grid], CTX)
        expected = torch.nn.functional.grid_sample(
            x, grid, mode="bilinear", padding_mode="border", align_corners=True
        )
        torch.testing.assert_close(result, expected)


class TestDepthToSpaceOp(unittest.TestCase):
    """Test DepthToSpace operator forward function."""

    def test_depth_to_space_dcr(self):
        op = make_op("d2s", "DepthToSpace", attributes={"blocksize": 2, "mode": "DCR"})
        x = torch.randn(1, 8, 2, 3)
        result = DEFAULT_BACKEND_TABLE["DepthToSpace"](op, [x], CTX)
        expected = torch.nn.functional.pixel_shuffle(x, 2)
        torch.testing.assert_close(result, expected)
        self.assertEqual(result.shape, (1, 2, 4, 6))

    def test_depth_to_space_crd(self):
        op = make_op("d2s_crd", "DepthToSpace", attributes={"blocksize": 2, "mode": "CRD"})
        x = torch.randn(1, 8, 2, 3)
        result = DEFAULT_BACKEND_TABLE["DepthToSpace"](op, [x], CTX)
        self.assertEqual(result.shape, (1, 2, 4, 6))
        # Verify CRD mode manually
        b, c, h, w = x.shape
        blocksize = 2
        tmp = x.reshape(b, c // (blocksize * blocksize), blocksize, blocksize, h, w)
        tmp = tmp.permute(0, 1, 4, 2, 5, 3)
        expected = tmp.reshape(b, c // (blocksize * blocksize), h * blocksize, w * blocksize)
        torch.testing.assert_close(result, expected)

    def test_depth_to_space_default_mode(self):
        op = make_op("d2s_def", "DepthToSpace", attributes={"blocksize": 2})
        x = torch.randn(1, 4, 3, 3)
        result = DEFAULT_BACKEND_TABLE["DepthToSpace"](op, [x], CTX)
        expected = torch.nn.functional.pixel_shuffle(x, 2)
        torch.testing.assert_close(result, expected)
        self.assertEqual(result.shape, (1, 1, 6, 6))


if __name__ == "__main__":
    unittest.main()



import torch


def torch_mean_square_error(y_pred: torch.Tensor, y_real: torch.Tensor, reduction: str = "mean") -> torch.Tensor:
    """
    Compute mean square error between y_pred(tensor) and y_real(tensor)

    MSE error can be calcualted as following equation:

        MSE(x, y) = (x - y) ^ 2

    if x and y are matrixs, MSE error over matrix should be the mean value of MSE error over all elements.

        MSE(X, Y) = mean((X - Y) ^ 2)

    By this equation, we can easily tell that MSE is an symmtrical measurement:
        MSE(X, Y) == MSE(Y, X)
        MSE(0, X) == X ^ 2

    Args:
        y_pred (torch.Tensor): _description_
        y_real (torch.Tensor): _description_
        reduction (str, optional): _description_. Defaults to 'mean'.

    Raises:
        ValueError: _description_
        ValueError: _description_

    Returns:
        torch.Tensor: _description_
    """
    if y_pred.shape != y_real.shape:
        raise ValueError(
            f"Can not compute mse loss for tensors with different shape. " f"({y_pred.shape} and {y_real.shape})"
        )
    reduction = str(reduction).lower()

    while y_pred.ndim < 2:
        y_pred = y_pred.unsqueeze(0)
        y_real = y_real.unsqueeze(0)

    diff = torch.pow(y_pred - y_real, 2).flatten(start_dim=1)
    mse = torch.mean(diff, dim=-1)

    if reduction == "mean":
        return torch.mean(mse)
    elif reduction == "sum":
        return torch.sum(mse)
    elif reduction == "none":
        return mse
    else:
        raise ValueError(f"Unsupported reduction method.")


def torch_snr_error(y_pred: torch.Tensor, y_real: torch.Tensor, reduction: str = "mean") -> torch.Tensor:
    """
    Compute SNR between y_pred(tensor) and y_real(tensor)

    SNR can be calcualted as following equation:

        SNR(pred, real) = (pred - real) ^ 2 / (real) ^ 2

    if x and y are matrixs, SNR error over matrix should be the mean value of SNR error over all elements.

        SNR(pred, real) = mean((pred - real) ^ 2 / (real) ^ 2)

    Args:
        y_pred (torch.Tensor): _description_
        y_real (torch.Tensor): _description_
        reduction (str, optional): _description_. Defaults to 'mean'.

    Raises:
        ValueError: _description_
        ValueError: _description_

    Returns:
        torch.Tensor: _description_
    """
    if y_pred.shape != y_real.shape:
        raise ValueError(
            f"Can not compute snr loss for tensors with different shape. " f"({y_pred.shape} and {y_real.shape})"
        )
    reduction = str(reduction).lower()

    while y_pred.ndim < 2:
        y_pred = y_pred.unsqueeze(0)
        y_real = y_real.unsqueeze(0)

    y_pred = y_pred.flatten(start_dim=1)
    y_real = y_real.flatten(start_dim=1)

    noise_power = torch.pow(y_pred - y_real, 2).sum(dim=-1)
    signal_power = torch.pow(y_real, 2).sum(dim=-1)
    snr = (noise_power) / (signal_power + 1e-7)

    if reduction == "mean":
        return torch.mean(snr)
    elif reduction == "sum":
        return torch.sum(snr)
    elif reduction == "none":
        return snr
    else:
        raise ValueError(f"Unsupported reduction method.")



from typing import Iterable, List, Tuple

import torch
from xslim.logger import logger

from ...executor import BaseGraphExecutor
from ...IR import BaseGraph, GraphFormatter, Operation, QuantableOperation, Variable
from ...IR.morph import GraphDecomposer
from .base import QuantizationOptimizationPass


class NXPResizeModeChangePass(QuantizationOptimizationPass):
    """This optimization pass overwrite resize mode to 'nearest' for all resize
    operations."""

    def __init__(self) -> None:
        super().__init__(name="NXP Resize Operation Transformation")

    def optimize(self, graph: BaseGraph, dataloader: Iterable, executor: BaseGraphExecutor, **kwargs) -> None:
        for op in graph.operations.values():
            if op.type == "Resize":
                op.attributes["mode"] = "nearest"
                op.attributes["coordinate_transformation_mode"] = "half_pixel"


class NCNNFormatGemmPass(QuantizationOptimizationPass):
    def __init__(self, name: str = "ncnn Format Gemm Pass") -> None:
        super().__init__(name)

    def optimize(self, graph: BaseGraph, dataloader: Iterable, executor: BaseGraphExecutor, **kwargs) -> None:
        for op in graph.operations.values():
            if op.type == "Gemm":
                if op.attributes.get("transB", 0) == 0:
                    op.attributes["transB"] = 1
                    weight = op.parameters[0].value
                    assert isinstance(weight, torch.Tensor)
                    op.parameters[0].value = weight.transpose(1, 0).contiguous()
                if op.attributes.get("alpha", 1.0) != 1.0:
                    op.parameters[0].value = op.parameters[0].value * op.attributes.get("alpha", 1.0)
                    op.attributes["alpha"] = 1.0
                if op.attributes.get("beta", 1.0) != 1.0:
                    if op.num_of_input > 2:
                        op.parameters[1].value = op.parameters[1].value * op.attributes.get("beta", 1.0)
                    op.attributes["beta"] = 1.0


class HorizontalLayerSplitPass(QuantizationOptimizationPass):
    """
    Horizontal Layer Split Pass(算子分裂过程)

    Split convolution layers or GEMM layers for better performance.

    Formula:

            Y = W * X + b

            where W can be divided into W_1 + W_2

            Y = (W_1 * X + b) + (W_2 * X)

    By splitting W like this, we are able to represent W more accurately.
    In the case where one channel has weights in the range [-32, 32] and another channel has weights in the range [-0.5, 0.5].
    the large channel will be divided so the range will come to [-16, 16], which leads us to use scale = 0.125 for representing
    the weight tensor rather than 0.25.

    The Estimation of Quantization Error is shown as a quadratic function of scale:

            E(Quantization Error) = scale ^ 2 / 12

    This Formula is proved by Bernard Widrow, according to the formula, a scale = 0.125 will decrease the quantization error by 75%.

    All the value larger than value_threshold will be divided into 2 part via this function, thus the layer itself will be
    splitted, an new Add operation are going to be created.

    ### Parameters:
        self.interested_layers = interested_layers
        self.value_threshold   = value_threshold
        self.method            = str(method).lower()
        self.verbose           = verbose

    # interested_layers(List[str])

            Only layer that listed in interested_layers will be processed by this pass.

            If interested_layers is None or empty list, NO layer will be processed.

    # value_threshold(float)

            This pass split value only when value is larger than value_threshold

            If there is no value large enough to be processed, corresponding layer will be skipped.

    # method(str)

            Splitting method, 'balance' or 'random'

            With balance method, W_1 and W_2 will be evenly divided.

            With random method, W_1 and W_2 will be randomly divided.

    ### Warning:

    Creating new operation in your network probably slows down the execution.

    Thus horizontal splitting is somehow a trade-off between speed and accuracy.

    ### Usage

    You can create this optimization manually:

        from ppq import HorizontalLayerSplitPass

        optim = HorizontalLayerSplitPass()
    """

    def __init__(
        self,
        interested_layers: List[str] = None,
        value_threshold: float = 1,
        method: str = "balance",
        verbose: bool = True,
    ) -> None:
        super().__init__("Layer Split Pass(Lateral)")
        self.interested_layers = interested_layers
        self.value_threshold = value_threshold
        self.method = str(method).lower()
        self.verbose = verbose

        if self.interested_layers is None or len(self.interested_layers) == 0:
            raise ValueError(
                "Layer Split Pass(Lateral) Requires a list of spliting layers, "
                "while parameter interested_layers is empty."
            )

        if self.method not in {"balance", "random"}:
            raise ValueError(f"Split method must be balance or random. While {self.method} is given.")

    def h_split(self, op: Operation) -> Tuple[torch.Tensor, torch.Tensor, int]:
        # split weight
        value = op.inputs[1].value
        mask = value.abs() > self.value_threshold
        processed_values = mask.sum().item()

        s_value = value
        if self.method == "balance":
            s_value = (value / 2) * mask
        elif self.method == "random":
            s_value = (value * torch.rand_like(value)) * mask
        else:
            raise Exception("Oops, seems we got some troubles here.")
        r_value = value - s_value

        # print
        if self.verbose:
            print("")
            print(f"# Layer {op.name} has been splited, " f"{processed_values}/{value.numel()} value(s) was processed.")
        return r_value, s_value, processed_values

    def optimize(self, graph: BaseGraph, dataloader: Iterable, executor: BaseGraphExecutor, **kwargs) -> None:
        with torch.no_grad():
            for name in self.interested_layers:
                # op check
                if name not in graph.operations:
                    raise KeyError(f"Operation {name} is not in current graph.")
                op1 = graph.operations[name]
                if op1.type not in {"Gemm", "MatMul", "Conv", "ConvTranspose"}:
                    raise TypeError(f"Operation {op1.name} can not be splited, op type is invalid({op1.type})")
                if not op1.inputs[1].is_parameter:
                    raise ValueError(f"Operation {op1.name} can not be splited, input 1 is not parameter.")
                if isinstance(op1, QuantableOperation):
                    raise TypeError(
                        f"Can not split a quantized operation, "
                        "Layer Split Pass should only be invoked as a pre-quant optimziation."
                    )

                r_value, s_value, processed_values = self.h_split(op1)

                if processed_values > 0:
                    # clone current operation
                    op2 = graph.create_operation(
                        op_type=op1.type, attributes=op1.attributes.copy(), platform=op1.platform
                    )
                    input_var, output_var = op1.inputs[0], op1.outputs[0]
                    graph.create_link_with_op(variable=op1.inputs[0], A=input_var.source_op, B=op2)

                    # create weight for cloned operation.
                    graph.create_variable(value=op1.inputs[1].value.clone(), is_parameter=True, dest_ops=[op2])

                    # set splited value
                    op1.inputs[1].value.copy_(r_value)
                    op2.inputs[1].value.copy_(s_value)

                    op1.outputs.clear()
                    adder = graph.create_operation(op_type="Add", platform=op1.platform, outputs=[output_var])
                    output_var.source_op = adder

                    graph.create_link_with_op(A=op1, B=adder)
                    graph.create_link_with_op(A=op2, B=adder)


class MetaxGemmSplitPass(QuantizationOptimizationPass):
    """Metax 不支持 Gemm 的量化，这个 pass 将 Gemm 拆分成.

    --- MatMul -----|
                    + --- Add ---
        bias   -----|
    """

    def __init__(self, name: str = "Metax Gemm Split Pass") -> None:
        super().__init__(name)

    # Implementation of Gemm Split will move to IR.morph soon.
    def optimize(self, graph: BaseGraph, dataloader: Iterable, executor: BaseGraphExecutor, **kwargs) -> None:
        morpher = GraphDecomposer(graph)
        morpher.decompose_gemm()


class GRUSplitPass(QuantizationOptimizationPass):
    """执行 GRU 算子分解，这个 Pass 将 GRU 算子分解为单步执行的形式.

    请注意，对于 ONNX GRU 算子而言, 它有两个输出, 一个是完整的hidden vector, 另一个是单步的 last state 这个
    pass 是针对单步执行而设计的，它将直接删除 hidden vector 之后的所有输出
    """

    def __init__(self, name: str = "Metax Gemm Split Pass") -> None:
        super().__init__(name)

    def delete_hidden_vec(self, graph: BaseGraph, hidden_vec: Variable):
        processor = GraphFormatter(graph)
        processor.truncate_on_var(var=hidden_vec, mark_as_output=False)

    # Implementation of Gemm Split will move to IR.morph soon.
    def optimize(self, graph: BaseGraph, dataloader: Iterable, executor: BaseGraphExecutor, **kwargs) -> None:
        interested_ops = []
        for operation in graph.operations.values():
            if operation.type == "GRU":
                interested_ops.append(operation)

        for op in interested_ops:
            assert isinstance(op, Operation)
            # fetch all related variables
            rnn_x, rnn_w, rnn_r, rnn_b, _, rnn_h = op.inputs
            hidden_size = op.attributes["hidden_size"]

            # Take a further look at
            # https://github.com/onnx/onnx/blob/main/docs/Operators.md#GRU
            Wz = rnn_w.value[0, hidden_size * 0 : hidden_size * 1]
            Wr = rnn_w.value[0, hidden_size * 1 : hidden_size * 2]
            Wh = rnn_w.value[0, hidden_size * 2 : hidden_size * 3]

            Rz = rnn_r.value[0, hidden_size * 0 : hidden_size * 1]
            Rr = rnn_r.value[0, hidden_size * 1 : hidden_size * 2]
            Rh = rnn_r.value[0, hidden_size * 2 : hidden_size * 3]

            Wbz = rnn_b.value[0, hidden_size * 0 : hidden_size * 1]
            Wbr = rnn_b.value[0, hidden_size * 1 : hidden_size * 2]
            Wbh = rnn_b.value[0, hidden_size * 2 : hidden_size * 3]

            Rbz = rnn_b.value[0, hidden_size * 3 : hidden_size * 4]
            Rbr = rnn_b.value[0, hidden_size * 4 : hidden_size * 5]
            Rbh = rnn_b.value[0, hidden_size * 5 : hidden_size * 6]

            # create operations
            op1 = graph.create_operation(op_type="Gemm", attributes={"transB": 1})
            op2 = graph.create_operation(op_type="Gemm", attributes={"transB": 1})
            op3 = graph.create_operation(op_type="Add")
            op4 = graph.create_operation(op_type="Sigmoid")
            op5 = graph.create_operation(op_type="Slice")
            op6 = graph.create_operation(op_type="Slice")
            op7 = graph.create_operation(op_type="Gemm", attributes={"transB": 1})
            op8 = graph.create_operation(op_type="Gemm", attributes={"transB": 1})
            op9 = graph.create_operation(op_type="Mul")
            op10 = graph.create_operation(op_type="Mul")
            op11 = graph.create_operation(op_type="Sub")
            op12 = graph.create_operation(op_type="Add")
            op13 = graph.create_operation(op_type="Mul")
            op14 = graph.create_operation(op_type="Tanh")
            op15 = graph.create_operation(op_type="Add")

            # create parameter variables
            # 为了加速运算，我们将Wz, Wr合并成Wzr, Rzh等同理
            # 参考 https://github.com/onnx/onnx/blob/main/docs/Operators.md#GRU
            Wzr_var = graph.create_variable(value=torch.cat([Wz, Wr]), is_parameter=True)
            Rzr_var = graph.create_variable(value=torch.cat([Rz, Rr]), is_parameter=True)
            Wbzr_var = graph.create_variable(value=torch.cat([Wbz, Wbr]), is_parameter=True)
            Rbzr_var = graph.create_variable(value=torch.cat([Rbz, Rbr]), is_parameter=True)

            Wh_var = graph.create_variable(value=Wh, is_parameter=True)
            Rh_var = graph.create_variable(value=Rh, is_parameter=True)
            Wbh_var = graph.create_variable(value=Wbh, is_parameter=True)
            Rbh_var = graph.create_variable(value=Rbh, is_parameter=True)

            constant_of_sub = graph.create_variable(value=torch.tensor(1.0).to(Wz.device), is_parameter=True)

            # link variables
            graph.create_link_with_op(variable=constant_of_sub, A=None, B=op11)
            graph.create_link_with_op(A=op1, B=op3)
            graph.create_link_with_op(A=op2, B=op3)
            graph.create_link_with_op(A=op3, B=op4)

            var = graph.create_variable()
            graph.create_link_with_op(variable=var, A=op4, B=op5)
            graph.create_link_with_op(variable=var, A=op4, B=op6)

            var = graph.create_variable()
            graph.create_link_with_op(variable=var, A=op5, B=op11)
            graph.create_link_with_op(variable=var, A=op5, B=op10)

            graph.create_link_with_op(A=op6, B=op9)
            graph.create_link_with_op(A=op7, B=op9)
            graph.create_link_with_op(A=op8, B=op12)
            graph.create_link_with_op(A=op9, B=op12)
            graph.create_link_with_op(A=op10, B=op15)
            graph.create_link_with_op(A=op11, B=op13)
            graph.create_link_with_op(A=op12, B=op14)
            graph.create_link_with_op(A=op13, B=op15)
            graph.create_link_with_op(A=op14, B=op13)

            # mark h as graph input, link h to op2, op10 and op7
            rnn_h.source_op.outputs.remove(rnn_h)
            rnn_h.source_op = None
            rnn_h.dest_ops.remove(op)
            graph.mark_variable_as_graph_input(rnn_h)
            graph.create_link_with_op(variable=rnn_h, A=None, B=op2)
            graph.create_link_with_op(variable=rnn_h, A=None, B=op7)
            graph.create_link_with_op(variable=rnn_h, A=None, B=op10)

            # link x to op1 and op8
            rnn_x.dest_ops.remove(op)
            graph.create_link_with_op(variable=rnn_x, A=rnn_x.source_op, B=op1)
            graph.create_link_with_op(variable=rnn_x, A=rnn_x.source_op, B=op8)

            # create parameters
            graph.create_link_with_op(variable=Wzr_var, A=None, B=op1)
            graph.create_link_with_op(variable=Rzr_var, A=None, B=op2)
            graph.create_link_with_op(variable=Wh_var, A=None, B=op8)
            graph.create_link_with_op(variable=Rh_var, A=None, B=op7)
            graph.create_link_with_op(variable=Wbzr_var, A=None, B=op1)
            graph.create_link_with_op(variable=Rbzr_var, A=None, B=op2)
            graph.create_link_with_op(variable=Wbh_var, A=None, B=op8)
            graph.create_link_with_op(variable=Rbh_var, A=None, B=op7)

            graph.create_link_with_op(
                variable=graph.create_variable(value=torch.tensor([0]), is_parameter=True), A=None, B=op5
            )
            graph.create_link_with_op(
                variable=graph.create_variable(value=torch.tensor([hidden_size]), is_parameter=True), A=None, B=op5
            )
            graph.create_link_with_op(
                variable=graph.create_variable(value=torch.tensor([1]), is_parameter=True), A=None, B=op5
            )
            graph.create_link_with_op(
                variable=graph.create_variable(value=torch.tensor([1]), is_parameter=True), A=None, B=op5
            )

            graph.create_link_with_op(
                variable=graph.create_variable(value=torch.tensor([hidden_size]), is_parameter=True), A=None, B=op6
            )
            graph.create_link_with_op(
                variable=graph.create_variable(value=torch.tensor([2 * hidden_size]), is_parameter=True), A=None, B=op6
            )
            graph.create_link_with_op(
                variable=graph.create_variable(value=torch.tensor([1]), is_parameter=True), A=None, B=op6
            )
            graph.create_link_with_op(
                variable=graph.create_variable(value=torch.tensor([1]), is_parameter=True), A=None, B=op6
            )

            hidden_vec, last_state = op.outputs
            last_state.source_op = op15
            op15.outputs.append(last_state)

            op.inputs.clear()
            op.outputs.clear()
            graph.remove_operation(op)
            self.delete_hidden_vec(graph, hidden_vec)



#!/usr/bin/env python3
# Copyright (c) 2023 SpacemiT. All rights reserved.
import functools
import math
import time
from typing import Callable, Dict, Iterable, List, Set, Tuple, Union

import numpy as np
import torch
from xslim.logger import logger

from ..defs import OBSERVER_FLOATING_MSE_FETCHES, OBSERVER_MIN_SCALE_THRESHOLD, OBSERVER_PERCENTILE
from ..ppq_decorator import (
    BaseTensorObserver,
    QuantizationProperty,
    QuantizationStates,
    TensorQuantizationConfig,
    Variable,
    minmax_to_scale_offset,
    ppq_common,
    ppq_observer,
    ppq_quant_param_computing_function,
    torch_KL_divergence,
)


class TorchXSlimObserver(BaseTensorObserver):
    def __init__(
        self,
        watch_on: Variable,
        quant_cfg: TensorQuantizationConfig,
        hist_bins: int = OBSERVER_FLOATING_MSE_FETCHES,
        single_alg: str = None,
    ):
        self._watch_on = watch_on
        self._phase = "Detecting Minmax"
        self._hist = None
        self._hist_scale = None
        self._min = None
        self._max = None
        self._hist_bins = hist_bins
        self._quant_cfg = quant_cfg
        if not ppq_common.OBSERVER_PERCENTILE_MANUL_OVERRIDE in quant_cfg.detail:
            self._percentile = OBSERVER_PERCENTILE
        else:
            self._percentile = quant_cfg.detail[ppq_common.OBSERVER_PERCENTILE_MANUL_OVERRIDE]

        if ppq_common.OBSERVER_MIN_SCALE_MANUL_OVERRIDE in quant_cfg.detail:
            self._scale_threshold = quant_cfg.detail[ppq_common.OBSERVER_MIN_SCALE_MANUL_OVERRIDE]
        else:
            self._scale_threshold = OBSERVER_MIN_SCALE_THRESHOLD

        self._single_alg = single_alg
        self._channel_min_max = []

        self._force_range_min = -(2**31)
        self._force_range_max = 2**31
        if watch_on.source_op is not None:
            self._force_range_min = (
                watch_on.source_op._detail.get("output_force_range", {})
                .get(watch_on.name, {})
                .get("min", self._force_range_min)
            )
            self._force_range_max = (
                watch_on.source_op._detail.get("output_force_range", {})
                .get(watch_on.name, {})
                .get("max", self._force_range_max)
            )
        self._percentile_collector = []
        self._min_val_collector = []
        self._max_val_collector = []
        self._none_value_detected = False
        self._value_device = "cpu"
        self._num_of_quant_levels = (self._quant_cfg.quant_max - self._quant_cfg.quant_min) + 1
        self._offset_step = int(self._hist_bins // self._num_of_quant_levels * 2)
        self._range_step = int(self._hist_bins // self._num_of_quant_levels * 2)
        observe_batch_size = self._hist_bins // self._range_step
        self._hist_p_tensor = torch.zeros(
            size=(observe_batch_size, self._hist_bins), dtype=torch.float, device=self._value_device
        )
        self._hist_q_tensor = torch.zeros(
            size=(observe_batch_size, self._hist_bins), dtype=torch.float, device=self._value_device
        )

    def observe(self, value: torch.Tensor):
        if self._quant_cfg.state not in {QuantizationStates.INITIAL}:
            return

        if isinstance(value, torch.Tensor):
            self._value_device = value.device.type

        if value is None or value.numel() == 0:
            self._none_value_detected = True
            return

        assert isinstance(value, torch.Tensor), "XSlimObserver can only deal with torch Tensor values"
        if self._phase == "Detecting Minmax":
            if self._quant_cfg.state == QuantizationStates.INITIAL:
                if self._quant_cfg.policy.has_property(QuantizationProperty.PER_TENSOR):
                    numel = value.numel()
                    min_idx, max_idx = int(numel * (1 - self._percentile)), int(numel * (self._percentile))
                    # torch.kthvalue needs index from 1 to numel ...
                    min_idx = max(0, min_idx) + 1
                    max_idx = min(max_idx, numel - 1) + 1
                    _min = torch.kthvalue(value.flatten(), k=min_idx, dim=0)[0].view(1, -1)
                    _max = torch.kthvalue(value.flatten(), k=max_idx, dim=0)[0].view(1, -1)
                    self._percentile_collector.append(torch.cat([_max, _min], dim=-1))
                    self._min_val_collector.append(value.min().reshape(1))
                    self._max_val_collector.append(value.max().reshape(1))
                elif self._quant_cfg.policy.has_property(QuantizationProperty.PER_CHANNEL):
                    raise PermissionError("Percentile observer can not deal with per channel quantization.")
                else:
                    raise TypeError("Min-max Observer only work with per-tensor or per-channel quantize policy.")

        elif self._phase == "Collating Hist":
            if self._hist is None:
                self._hist = torch.zeros(size=(self._hist_bins,), dtype=torch.int32, device=value.device)

            if self._quant_cfg.policy.has_property(QuantizationProperty.ASYMMETRICAL):
                hist = torch.histc(value, self._hist_bins, min=self._full_min_val, max=self._full_max_val)
                self._hist += hist.int()

            elif self._quant_cfg.policy.has_property(QuantizationProperty.SYMMETRICAL):
                hist = torch.histc(torch.abs(value), self._hist_bins, min=0, max=self._hist_scale * self._hist_bins)
                self._hist += hist.int()

            else:
                raise TypeError(
                    "Quantization Property is invalid, " "expect either ASYMMETRICAL or SYMMETRICAL config here."
                )

    def render_quantization_config(self):
        # If TQC is not prepared for calibration, just skip this execution.
        if self._quant_cfg.state not in {QuantizationStates.INITIAL}:
            return

        if self._none_value_detected:
            logger.warning("None value detected at var {}".format(self._watch_on.name))
            self._quant_cfg.scale = torch.tensor([1.0], dtype=torch.float32, device=self._value_device).squeeze(0)
            self._quant_cfg.offset = torch.tensor([0], dtype=torch.float32, device=self._value_device).squeeze(0)
            self._quant_cfg.state = QuantizationStates.ACTIVATED
            self._quant_cfg.detail["NONE_VALUE"] = True
            return

        if not self._quant_cfg.policy.has_property(QuantizationProperty.PER_TENSOR):
            raise ValueError("Hist observer can only apply with per-tensor quantization config.")

        if self._phase == "Detecting Minmax":
            self._full_min_val = torch.min(torch.cat(self._min_val_collector, dim=0)).cpu().item()
            self._full_max_val = torch.max(torch.cat(self._max_val_collector, dim=0)).cpu().item()
            percentile_reduce = torch.cat(self._percentile_collector, dim=0).float().mean(dim=0).cpu()

            self._percentile_min_val = percentile_reduce[1].item()
            self._percentile_max_val = percentile_reduce[0].item()

            if self._quant_cfg.policy.has_property(QuantizationProperty.SYMMETRICAL):
                hist_range = float(max(abs(self._full_max_val), abs(self._full_min_val)))
            else:
                hist_range = self._full_max_val - self._full_min_val

            self._hist_scale = hist_range / self._hist_bins
            self._phase = "Collating Hist"
        elif self._phase == "Collating Hist":
            scale, offset = self.hist_to_scale_offset(
                histogram=self._hist, hist_bins=self._hist_bins, hist_scale=self._hist_scale, config=self._quant_cfg
            )
            self._quant_cfg.scale = torch.tensor([scale], dtype=torch.float32, device=self._value_device).squeeze(0)
            self._quant_cfg.offset = torch.tensor([offset], dtype=torch.float32, device=self._value_device).squeeze(0)
            self._quant_cfg.state = QuantizationStates.ACTIVATED

    def compute_mse_loss(
        self, histogram: np.ndarray, bin_start: int, bin_range: int, num_of_quant_levels: int, hist_sum: float
    ) -> float:
        bin_end = bin_start + bin_range
        if bin_end > histogram.size:
            bin_end = histogram.size

        idx_range = np.arange(0, histogram.size).astype(np.float32)
        step = (bin_end - bin_start) / num_of_quant_levels
        if step < 1:
            step = 1

        mid_val = (step - 1) / 2
        idx_range[bin_start:bin_end] -= bin_start
        idx_range[bin_start:bin_end] = idx_range[bin_start:bin_end] % step
        idx_range[bin_start:bin_end] -= mid_val
        idx_range[bin_start:bin_end] = mid_val - np.abs(idx_range[bin_start:bin_end]) + 0.5

        idx_range[:bin_start] = bin_start - idx_range[:bin_start] - 0.5
        idx_range[bin_end:] = idx_range[bin_end:] - bin_end + 0.5

        histogram = histogram / hist_sum

        loss = (idx_range * idx_range * histogram).sum()

        return float(loss)

    def compute_kl_loss(
        self, histogram: torch.Tensor, bin_start: int, bin_range: int, num_of_quant_levels: int, hist_sum: float
    ) -> float:
        bin_end = bin_start + bin_range
        if bin_end > histogram.numel():
            bin_end = histogram.numel()
        bin_range = bin_end - bin_start

        p_hist = torch.zeros(size=(bin_range,), dtype=torch.float, device=histogram.device)
        p_hist[:bin_range].copy_(histogram[bin_start:bin_end])
        p_hist[bin_range - 1] += torch.sum(histogram[bin_end:])
        p_hist[0] += torch.sum(histogram[:bin_start])
        p_hist = p_hist / hist_sum

        expand_ratio = math.ceil(bin_range / num_of_quant_levels)
        zero_pad_num = expand_ratio * num_of_quant_levels - bin_range

        q_hist = torch.zeros(size=(bin_range + zero_pad_num,), dtype=torch.float, device=histogram.device)
        q_hist[:bin_range].copy_(histogram[bin_start:bin_end])
        q_hist = q_hist.reshape((num_of_quant_levels, expand_ratio))

        positive_map = q_hist > 0
        positive_cnt = positive_map.sum(axis=1, keepdim=True)
        positive_cnt[positive_cnt == 0] = 1
        q_hist = torch.div(q_hist.sum(axis=1, keepdim=True), positive_cnt)
        q_hist = q_hist.repeat([1, expand_ratio])
        q_hist = q_hist * positive_map
        q_hist = q_hist / torch.sum(q_hist)
        q_hist = q_hist.flatten()
        q_hist = q_hist[:bin_range]

        kl_loss = torch_KL_divergence(p_hist, q_hist)

        return float(kl_loss)

    def compute_mse_loss_batch(
        self,
        histogram: torch.Tensor,
        bin_start: int,
        bin_range: torch.Tensor,
        num_of_quant_levels: int,
        hist_sum: float,
    ) -> float:
        bin_end = bin_start + bin_range
        batch_size = bin_range.numel()
        idx_range = torch.range(0, histogram.numel() - 1).to(torch.int32).tile([batch_size, 1]).to(histogram.device)
        idx_range_cpy = (
            torch.range(0, histogram.numel() - 1).to(torch.float32).tile([batch_size, 1]).to(histogram.device)
        )
        step = (
            torch.clip(bin_range / num_of_quant_levels, 1, histogram.numel())
            .view(-1, 1)
            .to(torch.int32)
            .to(histogram.device)
        )
        mid_val = (step - 1) / 2

        idx_range[:, bin_start:] -= bin_start
        idx_range[:, bin_start:] = idx_range[:, bin_start:] % step
        idx_range = idx_range.to(torch.float32)
        idx_range[:, bin_start:] -= mid_val
        idx_range[:, bin_start:] = mid_val - torch.abs(idx_range[:, bin_start:]) + 0.5
        idx_range[:, :bin_start] = bin_start - idx_range[:, :bin_start] - 0.5

        for idx in range(batch_size):
            idx_range[idx, bin_end[idx] :].copy_(idx_range_cpy[idx, bin_end[idx] :] - bin_end[idx] + 0.5)

        histogram = histogram / hist_sum

        mse_loss = (idx_range * idx_range * histogram).sum(dim=-1)

        return mse_loss

    def compute_kl_loss_batch(
        self,
        histogram: torch.Tensor,
        bin_start: int,
        bin_range: torch.Tensor,
        num_of_quant_levels: int,
        hist_sum: float,
    ):
        batch_size = bin_range.numel()
        bin_end = bin_start + bin_range
        self._hist_p_tensor[:, : histogram.numel() - bin_start].copy_(histogram[bin_start:])
        self._hist_q_tensor[:, : histogram.numel() - bin_start].copy_(histogram[bin_start:])
        hist_start_sum = torch.sum(histogram[:bin_start])
        for idx in range(batch_size):
            self._hist_p_tensor[idx, bin_end[idx] - 1] += torch.sum(histogram[bin_end[idx] :])
            self._hist_p_tensor[idx, bin_end[idx] :] = 0
            self._hist_p_tensor[idx:, 0] += hist_start_sum
            self._hist_q_tensor[idx, bin_end[idx] :] = 0

        p_hist = self._hist_p_tensor[:batch_size] / hist_sum

        q_hist = self._hist_q_tensor[:batch_size].view(batch_size, num_of_quant_levels, -1)
        expand_size = q_hist.shape[-1]
        positive_map = q_hist > 0
        positive_cnt = positive_map.sum(axis=-1, keepdim=True)
        positive_cnt[positive_cnt == 0] = 1
        q_hist = torch.div(q_hist.sum(axis=-1, keepdim=True), positive_cnt)
        q_hist = q_hist.repeat([1, 1, expand_size])
        q_hist = q_hist * positive_map
        q_hist = q_hist.view(batch_size, -1)
        q_hist = q_hist / torch.sum(q_hist, dim=-1, keepdim=True)

        def torch_KL_divergence_batch(hist: torch.Tensor, ref_hist: torch.Tensor, eps=1e-30) -> float:
            if len(hist) != len(ref_hist):
                raise ValueError("Can not compute KL divergence, len(hist) != len(ref_hist")
            lhs = hist.double()
            batch_size = hist.shape[0]
            rhs = torch.log10(hist.double() + eps) - torch.log10(ref_hist.double() + eps)
            return torch.matmul(lhs.view(batch_size, 1, -1), rhs.view(batch_size, -1, 1)).view(-1)

        kl_loss = torch_KL_divergence_batch(p_hist, q_hist)

        return kl_loss

    @ppq_quant_param_computing_function
    def hist_to_scale_offset(
        self,
        histogram: torch.Tensor,
        hist_bins: int,
        hist_scale: float,
        config: TensorQuantizationConfig,
        scale_threshold: float = OBSERVER_MIN_SCALE_THRESHOLD,
    ) -> Tuple[float, int]:
        if config.policy.has_property(QuantizationProperty.ASYMMETRICAL) and config.policy.has_property(
            QuantizationProperty.PER_TENSOR
        ):
            t_start = time.time()

            self._hist_p_tensor = self._hist_p_tensor.to(histogram.device)
            self._hist_q_tensor = self._hist_q_tensor.to(histogram.device)
            histogram_cpu = histogram.float().cpu()
            histogram_np = histogram_cpu.numpy()

            losses = []

            percentile_scale, percentile_offset = minmax_to_scale_offset(
                max(self._force_range_min, self._percentile_min_val),
                min(self._force_range_max, self._percentile_max_val),
                config,
                self._scale_threshold,
            )

            full_scale, full_offset = minmax_to_scale_offset(
                max(self._force_range_min, self._full_min_val),
                min(self._force_range_max, self._full_max_val),
                config,
                self._scale_threshold,
            )

            if hist_scale == 0.0 or abs(self._percentile_min_val - self._percentile_max_val) < self._scale_threshold:
                logger.debug("observer render time cost {}".format(time.time() - t_start))
                return full_scale, full_offset

            percentile_bin_start = math.floor((self._percentile_min_val - self._full_min_val) / hist_scale)
            percentile_bin_start = max(percentile_bin_start, 0)
            percentile_bin_range = math.ceil((self._percentile_max_val - self._percentile_min_val) / hist_scale)
            if self._num_of_quant_levels > 2**10:
                logger.debug("observer render time cost {}".format(time.time() - t_start))
                return percentile_scale, percentile_offset

            hist_sum = float(histogram.sum())
            if percentile_bin_range + percentile_bin_start > hist_bins:
                percentile_bin_range = hist_bins - percentile_bin_start
            percentile_kl_loss = self.compute_kl_loss_batch(
                histogram,
                percentile_bin_start,
                torch.tensor([percentile_bin_range]),
                self._num_of_quant_levels,
                hist_sum,
            )[0].item()
            percentile_mse_loss = self.compute_mse_loss(
                histogram_np,
                percentile_bin_start,
                percentile_bin_range,
                self._num_of_quant_levels,
                hist_sum,
            )
            losses.append(
                {
                    "mse": percentile_mse_loss,
                    "kl": percentile_kl_loss,
                    "bin_range": percentile_bin_range,
                    "bin_start": percentile_bin_start,
                    "scale": percentile_scale,
                    "offset": percentile_offset,
                    "percentile_loss": True,
                }
            )

            for bin_start in range(0, hist_bins, self._offset_step):
                min_range_val_list = []
                max_range_val_list = []
                bin_range_list = []
                losses_start_idx = len(losses)
                for bin_range in range(
                    self._num_of_quant_levels,
                    hist_bins - bin_start + self._offset_step - 1,
                    self._range_step,
                ):
                    min_range_val = max(self._force_range_min, self._full_min_val + bin_start * hist_scale)
                    max_range_val = min(self._force_range_max, min_range_val + bin_range * hist_scale)
                    scale, offset = minmax_to_scale_offset(min_range_val, max_range_val, config, self._scale_threshold)
                    min_range_val_list.append(min_range_val)
                    max_range_val_list.append(max_range_val)
                    if bin_range + bin_start > hist_bins:
                        bin_range = hist_bins - bin_start
                    bin_range_list.append(bin_range)
                    mse_loss = self.compute_mse_loss(
                        histogram_np,
                        bin_start,
                        bin_range,
                        self._num_of_quant_levels,
                        hist_sum,
                    )
                    losses.append(
                        {
                            "mse": mse_loss,
                            "kl": 0,
                            "bin_range": bin_range,
                            "bin_start": bin_start,
                            "scale": scale,
                            "offset": offset,
                        }
                    )

                if len(bin_range_list) > 0:
                    kl_loss_tensor = self.compute_kl_loss_batch(
                        histogram,
                        bin_start,
                        torch.tensor(bin_range_list),
                        self._num_of_quant_levels,
                        hist_sum,
                    )

                    for idx, bin_range in enumerate(bin_range_list):
                        losses[losses_start_idx + idx]["kl"] = kl_loss_tensor[idx].item()

            losses_kl = sorted(losses, key=lambda x: x["kl"])
            losses_mse = sorted(losses_kl, key=lambda x: x["mse"])

            if self._single_alg == "mse":
                scale = losses_mse[0]["scale"]
                offset = losses_mse[0]["offset"]
            elif self._single_alg == "kl":
                scale = losses_kl[0]["scale"]
                offset = losses_kl[0]["offset"]
            else:
                valid_topk = 10
                loss_scale = [
                    x["kl"] * x["kl"] / (x["mse"] * x["mse"])
                    for x, y in zip(losses_mse[:valid_topk], losses_mse[:valid_topk])
                ]
                loss_scale = math.sqrt(sum(loss_scale) / len(loss_scale))
                losses = sorted(losses, key=lambda x: x["kl"] + x["mse"] * loss_scale)
                scale = losses[0]["scale"]
                offset = losses[0]["offset"]

            logger.debug("observer render time cost {}".format(time.time() - t_start))

            return scale, offset

        raise Exception("Oops, there might be some mistakes.")


class TorchXSlimMSEObserver(TorchXSlimObserver):
    def __init__(
        self,
        watch_on: Variable,
        quant_cfg: TensorQuantizationConfig,
        hist_bins: int = ppq_common.OBSERVER_FLOATING_MSE_FETCHES,
    ):
        super().__init__(watch_on, quant_cfg, hist_bins, "mse")


class TorchXSlimKLObserver(TorchXSlimObserver):
    def __init__(
        self,
        watch_on: Variable,
        quant_cfg: TensorQuantizationConfig,
        hist_bins: int = ppq_common.OBSERVER_FLOATING_MSE_FETCHES,
    ):
        super().__init__(watch_on, quant_cfg, hist_bins, "kl")


ppq_observer.OBSERVER_TABLE["kl"] = TorchXSlimKLObserver
ppq_observer.OBSERVER_TABLE["mse"] = TorchXSlimMSEObserver
ppq_observer.OBSERVER_TABLE["xslim"] = TorchXSlimObserver



from typing import Any, Union, Dict, Sequence
import torch
import cv2
import numpy as np


def preprocess_impl(path_list: Sequence[str], input_parametr: dict) -> torch.Tensor:
    """
    读取path_list, 并依据input_parametr中的参数预处理, 返回一个torch.Tensor

    Args:
        path_list (Sequence[str]): 一个校准batch的文件列表
        input_parametr (dict): 等同于配置中的calibration_parameters.input_parametres[idx]

    Returns:
        torch.Tensor: 一个batch的校准数据
    """
    # print("Custom Preprocessing ...")
    batch_list = []
    mean_value = input_parametr["mean_value"]
    std_value = input_parametr["std_value"]
    input_shape = input_parametr["input_shape"]
    for file_path in path_list:
        img = cv2.imread(file_path)
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        img = cv2.resize(img, (input_shape[-1], input_shape[-2]), interpolation=cv2.INTER_AREA)
        img = img.astype(np.float32)
        img = (img - mean_value) / std_value
        img = np.transpose(img, (2, 0, 1))
        img = torch.from_numpy(img)
        img = torch.unsqueeze(img, 0)
        batch_list.append(img)
    return torch.cat(batch_list, dim=0)



#!/usr/bin/env python3
# Copyright (c) 2023 SpacemiT. All rights reserved.
import copy
import json
import os
from enum import Enum
from typing import Dict, Literal, Pattern, Sequence, Union

import onnx
from xslim.logger import logger

from .defs import XQUANT_CONFIG, AutoFinetuneLevel, PrecisionLevel
from .ppq_decorator import BaseGraph


class SettingSerialize:
    def check(self, qsetting):
        pass

    def from_json(self, obj_setting: dict, qsetting):
        for key, value in obj_setting.items():
            if key in self.__dict__:
                if "builtin" in self.__dict__[key].__class__.__module__:
                    if isinstance(value, list) and hasattr(self, "from_list"):
                        self.__dict__[key] = getattr(self, "from_list")(key, value, qsetting)
                    else:
                        self.__dict__[key] = copy.deepcopy(value)
                elif isinstance(self.__dict__[key], Enum):
                    self.__dict__[key] = self.__dict__[key].__class__(value)
                else:
                    assert isinstance(value, dict)
                    if isinstance(self.__dict__[key], SettingSerialize):
                        self.__dict__[key].from_json(value, qsetting)

        self.check(qsetting)

    def to_json(self) -> str:
        return json.dumps(self, default=lambda o: o.__dict__, sort_keys=True, indent=4, ensure_ascii=False)

    def to_dict(self) -> str:
        return {k: v for k, v in self.__dict__.items()}


class ModelParameterSetting(SettingSerialize):
    def __init__(self) -> None:
        self.onnx_model: str = None
        self.output_prefix: str = None
        self.working_dir: str = None
        self.skip_onnxsim: bool = False

    def check(self, qsetting):
        # if not os.path.exists(self.onnx_model):
        #    raise FileExistsError(self.onnx_model)

        if self.working_dir is None:
            if isinstance(self.onnx_model, str) and os.path.exists(self.onnx_model):
                self.working_dir = os.path.dirname(self.onnx_model)
            else:
                self.working_dir = os.path.join(os.curdir, "temp")
            logger.info("Not set working_dir, deatults to {}.".format(self.working_dir))

        if self.output_prefix is None:
            if isinstance(self.onnx_model, str) and os.path.exists(self.onnx_model):
                model_name = os.path.splitext(os.path.basename(self.onnx_model))[0]
                self.output_prefix = "{}.q".format(model_name)
            else:
                self.output_prefix = "xslim.q"
            logger.info("Not set output_prefix, deatults to {}.".format(self.output_prefix))


class InputParameterSetting(SettingSerialize):
    def __init__(self) -> None:
        self.input_name: str = None
        self.input_shape: Sequence[int] = None
        self.file_type: Literal["img", "npy", "raw"] = "img"
        self.color_format: Literal["rgb", "bgr"] = "bgr"
        self.mean_value: Sequence[float] = None
        self.std_value: Sequence[float] = None
        self.preprocess_file: str = None
        self.data_list_path: str = None
        self.dtype: str = "float32"

    def check(self, qsetting):
        if self.preprocess_file is not None and not isinstance(self.preprocess_file, str):
            raise TypeError("preprocess_file type error, {} .vs str".format(self.preprocess_file))

        if self.data_list_path is not None and not isinstance(self.data_list_path, str):
            raise TypeError("data_list_path type error, {} .vs str".format(self.data_list_path))

        if isinstance(self.data_list_path, str) and not os.path.exists(self.data_list_path):
            raise FileExistsError(self.data_list_path)

        if self.mean_value is not None and not isinstance(self.mean_value, list):
            raise TypeError("mean_value type error, {} .vs str".format(self.mean_value))

        if self.std_value is not None and not isinstance(self.std_value, list):
            raise TypeError("std_value type error, {} .vs str".format(self.std_value))

        if self.file_type not in {"img", "npy", "raw"}:
            raise NotImplementedError("file_type {} not implemented yet.".format(self.file_type))


class CustomQuantizationParameterSetting(SettingSerialize):
    def __init__(self) -> None:
        self.input_names: Sequence[str] = None
        self.output_names: Sequence[str] = None
        self.max_percentile: float = None
        self.precision_level: int = None
        self.calibration_type: str = None


class QuantizationParameterSetting(SettingSerialize):
    def __init__(self) -> None:
        self.precision_level: PrecisionLevel = PrecisionLevel.LEVEL_0
        self.max_percentile: float = None
        self.finetune_level: AutoFinetuneLevel = AutoFinetuneLevel.LEVEL_1
        self.custom_setting: Sequence[CustomQuantizationParameterSetting] = None
        self.analysis_enable: bool = True
        self.truncate_var_names: Sequence[str] = []
        self.ignore_op_types: Sequence[str] = []
        self.ignore_op_names: Sequence[str] = []

    def from_list(self, value_name, obj_setting, qsetting):
        if value_name == "custom_setting":
            custom_setting = []
            for item_dict in obj_setting:
                setting_item = CustomQuantizationParameterSetting()
                setting_item.from_json(item_dict, qsetting)
                custom_setting.append(setting_item)
            return custom_setting
        else:
            return obj_setting

    def check(self, qsetting):
        if self.precision_level.value > PrecisionLevel.LEVEL_0.value:
            logger.info("set higher precision level.")
        if self.finetune_level.value > AutoFinetuneLevel.LEVEL_1.value:
            logger.info("set higher finetune level.")


class CalibrationParameterSetting(SettingSerialize):
    def __init__(self) -> None:
        self.calibration_batch_size: int = 1
        self.calibration_step: int = 500
        self.calibration_device: str = "cuda" if XQUANT_CONFIG.cuda_support else "cpu"
        self.calibration_type: str = "default"
        self.input_parametres: Sequence[InputParameterSetting] = None

    def from_list(self, value_name: str, obj_setting: Sequence, qsetting):
        if value_name == "input_parametres":
            input_parametres = []
            for item_dict in obj_setting:
                setting_item = InputParameterSetting()
                setting_item.from_json(item_dict, qsetting)
                input_parametres.append(setting_item)
            return input_parametres
        else:
            return obj_setting

    def check(self, qsetting):
        if self.calibration_device == "cuda" and not XQUANT_CONFIG.cuda_support:
            logger.warning("Specifies that cuda is used but not detected. Turn to cpu.")
            self.calibration_device = "cpu"

        if self.calibration_step > 1000 or self.calibration_step < 10:
            logger.warning("Specifies that calibration_step is too large or too small, it should be in [10, 1000].")

        assert len(self.input_parametres) > 0, "Calibration input_parametres setting not detected."

        if self.calibration_type not in {"default", "minmax", "percentile", "kl", "mse"}:
            raise NotImplementedError("calibration_type {} not implemented yet.".format(self.calibration_type))

        if self.calibration_type != "default":
            logger.info("set calibration_type {}.".format(self.calibration_type))

    def check_input_parametres(self, ppq_ir: BaseGraph):
        assert len(self.input_parametres) == len(
            ppq_ir.inputs
        ), "Calibration input_parametres size should <= model inputs size."

        pb_input = ppq_ir._detail.get("pb_input", [])
        for in_type, calib_parameter in zip(pb_input, self.input_parametres):
            input_shape = [
                i.dim_value if isinstance(i.dim_value, int) and i.dim_value > 0 else None
                for i in in_type.type.tensor_type.shape.dim
            ]
            if isinstance(in_type.type.tensor_type.elem_type, int):
                input_dtype = onnx.mapping.TENSOR_TYPE_TO_NP_TYPE[in_type.type.tensor_type.elem_type].name
            else:
                input_dtype = None
            if input_dtype is not None:
                calib_parameter.dtype = input_dtype
            if calib_parameter.input_name is None:
                calib_parameter.input_name = in_type.name
            if calib_parameter.input_shape is None and len(input_shape) > 0:
                input_shape[0] = input_shape[0] if isinstance(input_shape[0], int) else 1
                calib_parameter.input_shape = input_shape


class XSlimSetting(SettingSerialize):
    def __init__(self) -> None:
        self.model_parameters = ModelParameterSetting()
        self.calibration_parameters = CalibrationParameterSetting()
        self.quantization_parameters = QuantizationParameterSetting()


class XSlimSettingFactory:
    def from_json(json_obj: Union[str, dict]) -> XSlimSetting:
        if isinstance(json_obj, str):
            setting_dict = json.loads(json_obj)
        else:
            setting_dict = json_obj

        setting = XSlimSetting()
        setting.from_json(setting_dict, setting)

        return setting



import operator
import os
from functools import reduce
from typing import List

import torch
import torch.nn.functional as F
from torch import _VF


def ceil_div(a, b):
    return (a + b - 1) // b

def quantize_channel_blockwise(data, block=16, bit_size=4):
    M, K = data.shape
    align_k = ceil_div(K, block) * block
    pad_k = align_k - K
    if pad_k > 0:
        data = F.pad(data, [0, pad_k, 0, 0], "constant", 0.0)
    quant_range = (2 ** bit_size - 1) / 2
    quant_min = -2 ** (bit_size - 1)
    quant_max = 2 ** (bit_size - 1) - 1

    data_pack = data.reshape(M, align_k // block, block)
    data_pack = data_pack.permute(1, 0, 2).reshape(align_k // block, -1)
    data_pack_max = torch.max(torch.abs(data_pack), dim=1)[0]
    data_pack_scale = (data_pack_max / quant_range).reshape(-1, 1)
    data_pack_quant = torch.clip(torch.round(data_pack / data_pack_scale), quant_min, quant_max)
    data_pack_quant = data_pack_quant * data_pack_scale
    data_pack_quant = data_pack_quant.reshape(align_k // block, M, block).permute(1, 0, 2).reshape(M, align_k)
    if pad_k > 0:
        data_pack_quant = data_pack_quant[:, :K]
    return data_pack_quant

def quantize_tensorwise(data, block=16, bit_size=4):
    quant_range = (2 ** bit_size - 1) / 2
    quant_min = -2 ** (bit_size - 1)
    quant_max = 2 ** (bit_size - 1) - 1
    data_pack_max = torch.max(torch.abs(data))
    data_pack_scale = (data_pack_max / quant_range)
    data_pack_quant = torch.clip(torch.round(data / data_pack_scale), quant_min, quant_max)
    data_pack_quant = data_pack_quant * data_pack_scale
    return data_pack_quant

def quantize_blockwise(data, block=16, bit_size=4):
    # M, K
    M, K = data.shape
    align_m = ceil_div(M, block) * block
    align_k = ceil_div(K, block) * block
    pad_m = align_m - M
    pad_k = align_k - K
    if pad_m > 0 or pad_k > 0:
        data = F.pad(data, [0, pad_k, 0, pad_m], "constant", 0.0)
    quant_range = (2 ** bit_size - 1) / 2
    quant_min = -2 ** (bit_size - 1)
    quant_max = 2 ** (bit_size - 1) - 1
    data_pack = data.reshape(align_m // block, block, align_k // block, block)
    data_pack = data_pack.permute(0, 2, 1, 3).reshape(-1, block * block)
    data_pack_max = torch.max(torch.abs(data_pack), dim=1)[0]
    data_pack_scale = (data_pack_max / quant_range).reshape(-1, 1)
    data_pack_quant = torch.clip(torch.round(data_pack / data_pack_scale), quant_min, quant_max)
    data_pack_quant = data_pack_quant * data_pack_scale
    data_pack_quant = data_pack_quant.reshape(align_m // block, align_k // block, block, block).permute(0, 2, 1, 3).reshape(align_m, align_k)
    if pad_m > 0 or pad_k > 0:
        data_pack_quant = data_pack_quant[:M, :K]
    return data_pack_quant

def quantize_groupwise(data, group=16, bit_size=4):
    # M, K
    M, K = data.shape
    align_k = ceil_div(data.shape[1], group) * group
    pad_k = align_k - data.shape[1]
    if pad_k > 0:
        data = F.pad(data, [0, pad_k, 0, 0], "constant", 0.0)
    data_pack = data.reshape(-1, group)
    quant_range = (2 ** bit_size - 1) / 2
    quant_min = -2 ** (bit_size - 1)
    quant_max = 2 ** (bit_size - 1) - 1
    data_pack_max = torch.max(torch.abs(data_pack), dim=1)[0]
    data_pack_scale = (data_pack_max / quant_range).reshape(-1, 1)
    data_pack_quant = torch.clip(torch.round(data_pack / data_pack_scale), quant_min, quant_max)
    data_pack_quant = (data_pack_quant * data_pack_scale).reshape(-1, align_k)
    if pad_k > 0:
        data_pack_quant = data_pack_quant[:, :K]
    return data_pack_quant



from abc import ABCMeta, abstractproperty
from typing import Any, Dict, List

from ...core import (
    CAFFE_DOMAIN,
    COMPUTING_OP,
    DEFAULT_OPSET_DOMAIN,
    DEFAULT_OPSET_VERSION,
    ONNX_DOMAIN,
    OperationMeta,
    STRICT_OPSET_CHECKING,
    TargetPlatform,
)


class Opset:
    def __init__(self, domain: str = DEFAULT_OPSET_DOMAIN, version: int = DEFAULT_OPSET_VERSION) -> None:
        """Open Neural Network Exchange (ONNX) is an open ecosystem that empowers AI developers
            to choose the right tools as their project evolves.

        ONNX provides an open source format for AI models, both deep learning and traditional ML.
        It defines an extensible computation graph model, as well as definitions of
            built-in operators and standard data types.
        Currently we focus on the capabilities needed for inferencing (scoring).

        PPQ IR is built based on ONNX defination.

        Args:
            domain (str, optional): _description_. Defaults to DEFAULT_OPSET_DOMAIN.
            version (int, optional): _description_. Defaults to DEFAULT_OPSET_VERSION.
        """
        self.domain = domain
        self.version = version

    def is_onnx_v13(self):
        return self.domain == ONNX_DOMAIN and self.version == 13

    def is_onnx_v11(self):
        return self.domain == ONNX_DOMAIN and self.version == 11

    def onnx_opset_version(self) -> int:
        return self.version if self.domain == ONNX_DOMAIN else -1

    def is_onnx(self):
        return self.domain == ONNX_DOMAIN

    def is_caffe(self):
        return self.domain == CAFFE_DOMAIN


class OperationBase(metaclass=ABCMeta):
    def __init__(
        self,
        name: str,
        op_type: str,
        attributes: Dict[str, Any],
        opset=None,
        platform: TargetPlatform = TargetPlatform.UNSPECIFIED,
    ) -> None:
        self._name = name
        self._type = op_type
        self._attributes = attributes
        self._platform = platform
        self._meta = None
        self._input_vars = []
        self._output_vars = []
        self._detail = {}
        if opset is None:
            self._opset = Opset()
        else:
            self._opset = opset

    @abstractproperty
    def inputs(self) -> List[Any]:
        pass

    @abstractproperty
    def outputs(self) -> List[Any]:
        pass

    @abstractproperty
    def parameters(self) -> List[Any]:
        pass

    @property
    def name(self) -> str:
        return self._name

    @property
    def type(self) -> str:
        return self._type

    @type.setter
    def type(self, type: str):
        self._type = type

    @property
    def opset(self) -> Opset:
        return self._opset

    @opset.setter
    def opset(self, opset: Opset):
        self._opset = opset

    @property
    def attributes(self) -> Dict[str, Any]:
        return self._attributes

    @property
    def platform(self) -> TargetPlatform:
        return self._platform

    @platform.setter
    def platform(self, platform: TargetPlatform):
        self._platform = platform

    @property
    def num_of_input(self) -> int:
        return len(self._input_vars)

    @property
    def num_of_output(self) -> int:
        return len(self._output_vars)

    @property
    def meta_data(self) -> OperationMeta:
        raise NotImplementedError("This property has been removed since PPQ 0.6.6")

    @property
    def inputs(self) -> List[object]:
        return self._input_vars

    @property
    def outputs(self) -> List[object]:
        return self._output_vars

    @property
    def is_computing_op(self) -> bool:
        return self.type in COMPUTING_OP

    def set_extension_attrib(self, attrib: str, value: Any):
        self._detail[attrib] = value

    @property
    def extension_attrib(self) -> dict:
        return self._detail

    def __hash__(self) -> int:
        return self._name.__hash__()


class VLink:
    def __init__(self, in_idx: int, out_idx: int) -> None:
        if not isinstance(in_idx, int):
            raise TypeError(f"Can not create vlink with input_idx {in_idx}, " "only int value is acceptable here.")
        if not isinstance(out_idx, int):
            raise TypeError(f"Can not create vlink with output_idx {out_idx}, " "only int value is acceptable here.")
        self.in_idx = in_idx
        self.out_idx = out_idx


class OpSocket:
    def __init__(
        self,
        op: OperationBase,
        in_plat: List[TargetPlatform] = None,
        out_plat: List[TargetPlatform] = None,
        links: List[VLink] = None,
    ) -> None:
        self.in_plat = in_plat
        if in_plat is None:
            self.in_plat = [TargetPlatform.UNSPECIFIED for _ in range(op.num_of_input)]

        self.out_plat = out_plat
        if out_plat is None:
            self.out_plat = [TargetPlatform.UNSPECIFIED for _ in range(op.num_of_output)]

        self.links = links
        if self.links is None:
            self.links = []
            for i in range(op.num_of_input):
                for j in range(op.num_of_output):
                    self.links.append(VLink(i, j))


def CEHCK_TYPE(op: OperationBase, t: str):
    pass


def CHECK_OPSET(
    min_version_supported: int,
    max_version_supported: int,
    op: OperationBase,
    strict_check: bool = STRICT_OPSET_CHECKING,
):
    if not strict_check:
        return
    opset = op.opset
    if opset.domain != ONNX_DOMAIN:
        raise TypeError(
            f"Unrecognizable opset was found, "
            f"can not generate dispatching scheme with op {op.name}, "
            "cause it might not be a standard onnx operation."
        )

    if opset.version > max_version_supported or opset.version < min_version_supported:
        raise TypeError(
            f"opset version is not supported, "
            f"can not generate dispatching scheme with op {op.name}({op.type}), "
            f"currently we support only [{min_version_supported, max_version_supported}], "
            f"however {opset.version} was given."
        )


def DEFAULT_SOCKET_CREATOR(op: OperationBase) -> OpSocket:
    return OpSocket(op=op)


def Reshape_Socket(op: OperationBase) -> OpSocket:
    """
    From Opset 5 - 13:

    Inputs
        data (differentiable) : T
            An input tensor.

        shape (non-differentiable) : tensor(int64)
            Specified shape for output.

    Outputs
        reshaped (differentiable) : T
            Reshaped data.
    """
    CEHCK_TYPE(op=op, t="Reshape")
    CHECK_OPSET(op=op, min_version_supported=5, max_version_supported=13)
    return OpSocket(op=op, in_plat=[TargetPlatform.UNSPECIFIED, TargetPlatform.SOI], links=[VLink(in_idx=0, out_idx=0)])


def Pad_Socket(op: OperationBase) -> OpSocket:
    """
    From Opset 1 - 13:

    Inputs (2 - 3)
        data (differentiable) : T
            Input tensor.

        pads (non-differentiable) : tensor(int64)
            Tensor of integers indicating the number of padding elements to add or remove
            (if negative) at the beginning and end of each axis. For 2D input tensor, it is the number of pixels.
            `pads` should be a 1D tensor of shape [2 * input_rank].
            `pads` format should be: [x1_begin, x2_begin,...,x1_end, x2_end,...],
            where xi_begin is the number of pad values added at the beginning of axis `i` and xi_end,
            the number of pad values added at the end of axis `i`.

        constant_value (optional, non-differentiable) : T
            (Optional) A scalar value to be used if the mode chosen is `constant`
            (by default it is 0, empty string or False).

    Outputs
        output (differentiable) : T
            Tensor after padding.
    """
    CEHCK_TYPE(op=op, t="Pad")
    CHECK_OPSET(op=op, min_version_supported=1, max_version_supported=13)
    in_plat = [TargetPlatform.UNSPECIFIED, TargetPlatform.SOI, TargetPlatform.SOI]
    return OpSocket(op=op, in_plat=in_plat[: op.num_of_input], links=[VLink(in_idx=0, out_idx=0)])


def Gather_Socket(op: OperationBase) -> OpSocket:
    """
    From Opset 1 - 13:

    Inputs
        data (differentiable) : T
            Tensor of rank r >= 1.

        indices (non-differentiable) : Tind
            Tensor of int32/int64 indices, of any rank q.
            All index values are expected to be within bounds [-s, s-1] along axis of size s.
            It is an error if any of the index values are out of bounds.

    Outputs
        output (differentiable) : T
            Tensor of rank q + (r - 1).
    """
    CEHCK_TYPE(op=op, t="Gather")
    CHECK_OPSET(op=op, min_version_supported=1, max_version_supported=13)
    return OpSocket(
        op=op, in_plat=[TargetPlatform.UNSPECIFIED, TargetPlatform.FP32], links=[VLink(in_idx=0, out_idx=0)]
    )


def Resize_Socket(op: OperationBase) -> OpSocket:
    """
    From Opset 11 - 13:

    Inputs (1 - 4)
        X (differentiable) : T1
            N-D tensor

        roi (optional, non-differentiable) : T2
            1-D tensor given as [start1, ..., startN, end1, ..., endN],
            where N is the rank of X. The RoIs' coordinates are normalized in the coordinate system of the input image.
            It only takes effect when coordinate_transformation_mode is "tf_crop_and_resize"

        scales (optional, non-differentiable) : tensor(float)
            The scale array along each dimension. It takes value greater than 0.
            If it's less than 1, it's sampling down, otherwise, it's upsampling.
            The number of elements of 'scales' should be the same as the rank of input 'X'.
            One of 'scales' and 'sizes' MUST be specified and it is an error if both are specified.
            If 'sizes' is needed, the user can use an empty string as the name of 'scales' in this operator's input list.

        sizes (optional, non-differentiable) : tensor(int64)
            The size of the output tensor. The number of elements of 'sizes' should be the same as the rank of input 'X'.
            Only one of 'scales' and 'sizes' can be specified.

    Outputs
        Y (differentiable) : T1
            N-D tensor after resizing
    """
    CEHCK_TYPE(op=op, t="Resize")
    CHECK_OPSET(op=op, min_version_supported=10, max_version_supported=19, strict_check=True)
    in_plat = [TargetPlatform.UNSPECIFIED, TargetPlatform.SOI, TargetPlatform.SOI, TargetPlatform.SOI]
    return OpSocket(op=op, in_plat=in_plat[: op.num_of_input], links=[VLink(in_idx=0, out_idx=0)])


def Split_Socket(op: OperationBase) -> OpSocket:
    """
    From Opset 1 - 13:

    Inputs (1 - 2)
        input (differentiable) : T
            The tensor to split

        split (optional, non-differentiable) : tensor(int64)
            Optional length of each output. Values should be >= 0.Sum of the values
            must be equal to the dim value at 'axis' specified.

    Outputs (1 - ∞)
        outputs (variadic, differentiable) : T
            One or more outputs forming list of tensors after splitting
    """
    CEHCK_TYPE(op=op, t="Split")
    CHECK_OPSET(op=op, min_version_supported=1, max_version_supported=13)
    in_plat = [TargetPlatform.UNSPECIFIED, TargetPlatform.SOI]
    return OpSocket(op=op, in_plat=in_plat[: op.num_of_input], links=[VLink(in_idx=0, out_idx=0)])


def Topk_Socket(op: OperationBase) -> OpSocket:
    """
    From Opset 1 - 11:

    Inputs
        X (differentiable) : T
            Tensor of shape [a_1, a_2, ..., a_n, r]

        K (non-differentiable) : tensor(int64)
            A 1-D tensor containing a single positive value corresponding to the number of top elements to retrieve

    Outputs
        Values (differentiable) : T
            Tensor of shape [a_1, a_2, ..., a_{axis-1}, k, a_{axis+1}, ... a_n]
            containing top K values from the input tensor

        Indices (non-differentiable) : I
            Tensor of shape [a_1, a_2, ..., a_{axis-1}, k, a_{axis+1}, ... a_n]
            containing the corresponding input tensor indices for the top K values.
    """
    CEHCK_TYPE(op=op, t="TopK")
    CHECK_OPSET(op=op, min_version_supported=1, max_version_supported=11)
    in_plat = [TargetPlatform.UNSPECIFIED, TargetPlatform.SOI]
    out_plat = [TargetPlatform.UNSPECIFIED, TargetPlatform.FP32]
    return OpSocket(op=op, in_plat=in_plat[: op.num_of_input], out_plat=out_plat, links=[VLink(in_idx=0, out_idx=0)])


def Tile_Socket(op: OperationBase) -> OpSocket:
    """
    From Opset 1 - 13:

    Inputs
        input (differentiable) : T
            Input tensor of any shape.

    repeats (non-differentiable) : T1
        1D int64 tensor of the same length as input's dimension number,
        includes numbers of repeated copies along input's dimensions.

    Outputs
        output (differentiable) : T
            Output tensor of the same dimensions and type as tensor input.
            output_dim[i] = input_dim[i] * repeats[i]
    """
    CEHCK_TYPE(op=op, t="Tile")
    CHECK_OPSET(op=op, min_version_supported=1, max_version_supported=13)
    in_plat = [TargetPlatform.UNSPECIFIED, TargetPlatform.SOI, TargetPlatform.SOI]
    return OpSocket(op=op, in_plat=in_plat[: op.num_of_input], links=[VLink(in_idx=0, out_idx=0)])


def Expand_Socket(op: OperationBase) -> OpSocket:
    """
    From Opset 8 - 13:

    Inputs
        input (differentiable) : T

    Input tensor
        shape (non-differentiable) : tensor(int64)
            A 1-D tensor indicates the shape you want to expand to, following the broadcast rule

    Outputs
        output (differentiable) : T
            Output tensor
    """
    CEHCK_TYPE(op=op, t="Expand")
    CHECK_OPSET(op=op, min_version_supported=8, max_version_supported=13)
    in_plat = [TargetPlatform.UNSPECIFIED, TargetPlatform.SOI]
    return OpSocket(op=op, in_plat=in_plat[: op.num_of_input], links=[VLink(in_idx=0, out_idx=0)])


def RoiAlign_Socket(op: OperationBase) -> OpSocket:
    """
    From Opset 10 - 16:

    Inputs
        X : T1
            Input data tensor from the previous operator; 4-D feature map of shape (N, C, H, W),
            where N is the batch size, C is the number of channels,
            and H and W are the height and the width of the data.

        rois : T1
            RoIs (Regions of Interest) to pool over; rois is 2-D input of shape (num_rois, 4)
            given as [[x1, y1, x2, y2], ...]. The RoIs' coordinates are in the coordinate system of the input image.
            Each coordinate set has a 1:1 correspondence with the 'batch_indices' input.

        batch_indices : T2
            1-D tensor of shape (num_rois,) with each element denoting the index of the corresponding image in the batch.

    Outputs
        Y : T1
            RoI pooled output, 4-D tensor of shape (num_rois, C, output_height, output_width).
            The r-th batch element Y[r-1] is a pooled feature map corresponding to the r-th RoI X[r-1].
    """
    CEHCK_TYPE(op=op, t="RoiAlign")
    CHECK_OPSET(op=op, min_version_supported=10, max_version_supported=16)
    in_plat = [TargetPlatform.UNSPECIFIED, TargetPlatform.FP32, TargetPlatform.SOI]
    return OpSocket(op=op, in_plat=in_plat[: op.num_of_input], links=[VLink(in_idx=0, out_idx=0)])


def GridSampler_Socket(op: OperationBase) -> OpSocket:
    """
    From MMCV
    """
    in_plat = [TargetPlatform.UNSPECIFIED, TargetPlatform.FP32]
    return OpSocket(op=op, in_plat=in_plat[: op.num_of_input], links=[VLink(in_idx=0, out_idx=0)])


def Clip_Socket(op: OperationBase) -> OpSocket:
    """
    From Opset 1 - 13:

    Inputs (1 - 3)
        input (differentiable) : T
            Input tensor whose elements to be clipped

        min (optional, non-differentiable) : T
            Minimum value, under which element is replaced by min.
            It must be a scalar(tensor of empty shape).

        max (optional, non-differentiable) : T
            Maximum value, above which element is replaced by max.
            It must be a scalar(tensor of empty shape).

    Outputs
        output (differentiable) : T
            Output tensor with clipped input elements
    """
    CEHCK_TYPE(op=op, t="Clip")
    CHECK_OPSET(op=op, min_version_supported=1, max_version_supported=13)
    in_plat = [TargetPlatform.UNSPECIFIED, TargetPlatform.FP32, TargetPlatform.FP32]
    return OpSocket(op=op, in_plat=in_plat[: op.num_of_input], links=[VLink(in_idx=0, out_idx=0)])


def ConstantOfShape_Socket(op: OperationBase) -> OpSocket:
    """
    From Opset 9 - 9:

    Inputs
        input : T1
            1D tensor. The shape of the expected output tensor.
            If empty tensor is given, the output would be a scalar. All values must be >= 0.

    Outputs
        output : T2
            Output tensor of shape specified by 'input'.If attribute 'value' is specified,
            the value and datatype of the output tensor is taken from 'value'.If attribute 'value' is not specified,
            the value in the output defaults to 0, and the datatype defaults to float32.
    """
    CEHCK_TYPE(op=op, t="ConstantOfShape")
    CHECK_OPSET(op=op, min_version_supported=1, max_version_supported=13)
    in_plat = [TargetPlatform.SOI]
    return OpSocket(op=op, in_plat=in_plat[: op.num_of_input], links=[])


def GatherElements_Socket(op: OperationBase) -> OpSocket:
    """
    From Opset 11 - 13:

    Inputs
        data (differentiable) : T
            Tensor of rank r >= 1.

        indices (non-differentiable) : Tind
            Tensor of int32/int64 indices, with the same rank r as the input.
            All index values are expected to be within bounds [-s, s-1] along axis of size s.
            It is an error if any of the index values are out of bounds.

    Outputs
        output (differentiable) : T
            Tensor of the same shape as indices.
    """
    CEHCK_TYPE(op=op, t="GatherElements")
    CHECK_OPSET(op=op, min_version_supported=11, max_version_supported=13)
    in_plat = [TargetPlatform.UNSPECIFIED, TargetPlatform.FP32]
    return OpSocket(op=op, in_plat=in_plat[: op.num_of_input], links=[VLink(in_idx=0, out_idx=0)])


def GatherND_Socket(op: OperationBase) -> OpSocket:
    """
    From Opset 11 - 13:

    Inputs
        data (differentiable) : T
            Tensor of rank r >= 1.

        indices (non-differentiable) : tensor(int64)
            Tensor of rank q >= 1. All index values are expected to be within bounds [-s, s-1] along axis of size s.
            It is an error if any of the index values are out of bounds.

    Outputs
        output (differentiable) : T
            Tensor of rank q + r - indices_shape[-1] - 1.
    """
    CEHCK_TYPE(op=op, t="GatherND")
    CHECK_OPSET(op=op, min_version_supported=11, max_version_supported=13)
    in_plat = [TargetPlatform.UNSPECIFIED, TargetPlatform.FP32]
    return OpSocket(op=op, in_plat=in_plat[: op.num_of_input], links=[VLink(in_idx=0, out_idx=0)])


def NonMaxSuppression_Socket(op: OperationBase) -> OpSocket:
    """
    From Opset 10 - 13:

    Inputs (2 - 5)
        boxes : tensor(float)
            An input tensor with shape [num_batches, spatial_dimension, 4].
            The single box data format is indicated by center_point_box.

        scores : tensor(float)
            An input tensor with shape [num_batches, num_classes, spatial_dimension]

        max_output_boxes_per_class (optional) : tensor(int64)
            Integer representing the maximum number of boxes to be selected per batch per class.
            It is a scalar. Default to 0, which means no output.

        iou_threshold (optional) : tensor(float)
            Float representing the threshold for deciding whether boxes overlap too much with respect to IOU.
            It is scalar. Value range [0, 1]. Default to 0.

        score_threshold (optional) : tensor(float)
            Float representing the threshold for deciding when to remove boxes based on score. It is a scalar.

    Outputs
        selected_indices : tensor(int64)
            selected indices from the boxes tensor. [num_selected_indices, 3],
            the selected index format is [batch_index, class_index, box_index].
    """
    CEHCK_TYPE(op=op, t="NonMaxSuppression")
    CHECK_OPSET(op=op, min_version_supported=10, max_version_supported=13)
    in_plat = [
        TargetPlatform.UNSPECIFIED,
        TargetPlatform.FP32,
        TargetPlatform.SOI,
        TargetPlatform.SOI,
        TargetPlatform.SOI,
    ]
    return OpSocket(op=op, in_plat=in_plat[: op.num_of_input], links=[])


def NonZero_Socket(op: OperationBase) -> OpSocket:
    """
    From Opset 9 - 13:

    Inputs
        X (non-differentiable) : T
            input

    Outputs
        Y (non-differentiable) : tensor(int64)
            output
    """
    CHECK_OPSET(op=op, min_version_supported=9, max_version_supported=13)
    in_plat = [TargetPlatform.UNSPECIFIED]
    out_plat = [TargetPlatform.FP32]
    return OpSocket(op=op, in_plat=in_plat[: op.num_of_input], out_plat=out_plat, links=[])


def Range_Socket(op: OperationBase) -> OpSocket:
    """
    From Opset 11 - 13:

    Inputs
        start : T
            Scalar. First entry for the range of output values.

        limit : T
            Scalar. Exclusive upper limit for the range of output values.

        delta : T
            Scalar. Value to step by.

    Outputs
        output : T
            A 1-D tensor with same type as the inputs containing generated range of values.
    """
    CHECK_OPSET(op=op, min_version_supported=9, max_version_supported=13)
    in_plat = [TargetPlatform.SOI, TargetPlatform.SOI, TargetPlatform.SOI]
    return OpSocket(op=op, in_plat=in_plat[: op.num_of_input], links=[])


def ScatterElements_Socket(op: OperationBase) -> OpSocket:
    """
    From Opset 11 - 16:

    Inputs
        data (differentiable) : T
            Tensor of rank r >= 1.

        indices (non-differentiable) : Tind
            Tensor of int32/int64 indices, of r >= 1 (same rank as input).
            All index values are expected to be within bounds [-s, s-1] along axis of size s.
            It is an error if any of the index values are out of bounds.

        updates (differentiable) : T
            Tensor of rank r >=1 (same rank and shape as indices)

    Outputs
        output (differentiable) : T
            Tensor of rank r >= 1 (same rank as input).
    """
    CHECK_OPSET(op=op, min_version_supported=11, max_version_supported=16)
    in_plat = [TargetPlatform.UNSPECIFIED, TargetPlatform.FP32, TargetPlatform.UNSPECIFIED]
    return OpSocket(
        op=op, in_plat=in_plat[: op.num_of_input], links=[VLink(in_idx=0, out_idx=0), VLink(in_idx=2, out_idx=0)]
    )


def ScatterND_Socket(op: OperationBase) -> OpSocket:
    """
    From Opset 11 - 16:

    Inputs
        data (differentiable) : T
            Tensor of rank r >= 1.

        indices (non-differentiable) : tensor(int64)
            Tensor of rank q >= 1.

        updates (differentiable) : T
            Tensor of rank q + r - indices_shape[-1] - 1.

    Outputs
        output (differentiable) : T
            Tensor of rank r >= 1.
    """
    CHECK_OPSET(op=op, min_version_supported=11, max_version_supported=16)
    in_plat = [TargetPlatform.UNSPECIFIED, TargetPlatform.FP32, TargetPlatform.UNSPECIFIED]
    return OpSocket(
        op=op, in_plat=in_plat[: op.num_of_input], links=[VLink(in_idx=0, out_idx=0), VLink(in_idx=2, out_idx=0)]
    )


def Shape_Socket(op: OperationBase) -> OpSocket:
    """
    From Opset 1 - 16:

    Inputs
        data (non-differentiable) : T
            An input tensor.

    Outputs
        shape (non-differentiable) : T1
            Shape of the input tensor
    """
    CHECK_OPSET(op=op, min_version_supported=1, max_version_supported=16)
    in_plat = [TargetPlatform.UNSPECIFIED]
    out_plat = [TargetPlatform.SOI]
    return OpSocket(op=op, in_plat=in_plat[: op.num_of_input], out_plat=out_plat, links=[])


def Slice_Socket(op: OperationBase) -> OpSocket:
    """
    From Opset 1 - 16:

    Inputs (3 - 5)
        data : T
            Tensor of data to extract slices from.

        starts : Tind
            1-D tensor of starting indices of corresponding axis in `axes`

        ends : Tind
            1-D tensor of ending indices (exclusive) of corresponding axis in `axes`

        axes (optional) : Tind
            1-D tensor of axes that `starts` and `ends` apply to.
            Negative value means counting dimensions from the back. Accepted range is [-r, r-1] where r = rank(data).

        steps (optional) : Tind
            1-D tensor of slice step of corresponding axis in `axes`.
            Negative value means slicing backward. 'steps' cannot be 0. Defaults to 1.

    Outputs
        output : T
            Sliced data tensor.
    """
    CHECK_OPSET(op=op, min_version_supported=1, max_version_supported=13)
    in_plat = [
        TargetPlatform.UNSPECIFIED,
        TargetPlatform.SOI,
        TargetPlatform.SOI,
        TargetPlatform.SOI,
        TargetPlatform.SOI,
    ]
    return OpSocket(op=op, in_plat=in_plat[: op.num_of_input], links=[VLink(in_idx=0, out_idx=0)])


def Squeeze_Socket(op: OperationBase) -> OpSocket:
    """
    From Opset 1 - 13:

    Inputs (1 - 2)
        data (differentiable) : T
            Tensors with at least max(dims) dimensions.

        axes (optional, non-differentiable) : tensor(int64)
            List of integers indicating the dimensions to squeeze.
            Negative value means counting dimensions from the back.
            Accepted range is [-r, r-1] where r = rank(data).

    Outputs
        squeezed (differentiable) : T
            Reshaped tensor with same data as input.
    """
    CHECK_OPSET(op=op, min_version_supported=1, max_version_supported=13)
    in_plat = [TargetPlatform.UNSPECIFIED, TargetPlatform.SOI]
    return OpSocket(op=op, in_plat=in_plat[: op.num_of_input], links=[VLink(in_idx=0, out_idx=0)])


def Unsqueeze_Socket(op: OperationBase) -> OpSocket:
    """
    From Opset 1 - 13:

    Inputs
        data (differentiable) : T
            Original tensor

        axes (non-differentiable) : tensor(int64)
            List of integers indicating the dimensions to be inserted.
            Negative value means counting dimensions from the back.
            Accepted range is [-r, r-1] where r = rank(expanded).

    Outputs
        expanded (differentiable) : T
            Reshaped tensor with same data as input.
    """
    CHECK_OPSET(op=op, min_version_supported=1, max_version_supported=13)
    in_plat = [TargetPlatform.UNSPECIFIED, TargetPlatform.SOI]
    return OpSocket(op=op, in_plat=in_plat[: op.num_of_input], links=[VLink(in_idx=0, out_idx=0)])


def Where_Socket(op: OperationBase) -> OpSocket:
    """
    From Opset 9 - 16:

    Inputs
        condition (non-differentiable) : B
            When True (nonzero), yield X, otherwise yield Y

        X (differentiable) : T
            values selected at indices where condition is True

        Y (differentiable) : T
            values selected at indices where condition is False

    Outputs
        output (differentiable) : T
            Tensor of shape equal to the broadcasted shape of condition, X, and Y.
    """
    CHECK_OPSET(op=op, min_version_supported=9, max_version_supported=16)
    in_plat = [TargetPlatform.UNSPECIFIED, TargetPlatform.FP32, TargetPlatform.FP32]
    return OpSocket(op=op, in_plat=in_plat[: op.num_of_input], links=[VLink(in_idx=0, out_idx=0)])


def Logical_Socket(op: OperationBase) -> OpSocket:
    """
    From Opset 1 - 16:

    Inputs
        A (non-differentiable) : T
            First input operand for the logical operator.

        B (non-differentiable) : T
            Second input operand for the logical operator.

    Outputs
        C (non-differentiable) : T1
            Result tensor.
    """
    CHECK_OPSET(op=op, min_version_supported=12, max_version_supported=16)
    in_plat = [TargetPlatform.UNSPECIFIED, TargetPlatform.UNSPECIFIED]
    out_plat = [TargetPlatform.FP32]
    return OpSocket(op=op, in_plat=in_plat[: op.num_of_input], out_plat=out_plat, links=[])


def Onehot_Socket(op: OperationBase) -> OpSocket:
    """
    Inputs
        indices (non-differentiable) : T1

        depth (non-differentiable) : T2

        values (non-differentiable) : T3

    Outputs
        output (non-differentiable) : T3

    Args:
        op (Operation): _description_

    Returns:
        OpSocket: _description_
    """
    CHECK_OPSET(op=op, min_version_supported=12, max_version_supported=16)
    in_plat = [TargetPlatform.SOI, TargetPlatform.SOI, TargetPlatform.SOI]
    out_plat = [TargetPlatform.SOI]
    return OpSocket(op=op, in_plat=in_plat[: op.num_of_input], out_plat=out_plat, links=[])


DEFAULT_SOCKET_TABLE = {
    "AdaptiveAvgPool2d": DEFAULT_SOCKET_CREATOR,
    "Add": DEFAULT_SOCKET_CREATOR,
    "ArgMax": DEFAULT_SOCKET_CREATOR,
    "AveragePool": DEFAULT_SOCKET_CREATOR,
    "BatchNormalization": DEFAULT_SOCKET_CREATOR,
    "Cast": DEFAULT_SOCKET_CREATOR,
    "Clip": Clip_Socket,
    "Concat": DEFAULT_SOCKET_CREATOR,
    "Constant": DEFAULT_SOCKET_CREATOR,
    "ConstantOfShape": ConstantOfShape_Socket,
    "Conv": DEFAULT_SOCKET_CREATOR,
    "ConvTranspose": DEFAULT_SOCKET_CREATOR,
    "Div": DEFAULT_SOCKET_CREATOR,
    "Equal": Logical_Socket,
    "Exp": DEFAULT_SOCKET_CREATOR,
    "Expand": Expand_Socket,
    "Flatten": DEFAULT_SOCKET_CREATOR,
    "Gather": Gather_Socket,
    "GatherElements": GatherElements_Socket,
    "GatherND": GatherND_Socket,
    "Gelu": DEFAULT_SOCKET_CREATOR,
    "Gemm": DEFAULT_SOCKET_CREATOR,
    "grid_sampler": GridSampler_Socket,
    "GridSample": GridSampler_Socket,
    "GlobalAveragePool": DEFAULT_SOCKET_CREATOR,
    "GlobalMaxPool": DEFAULT_SOCKET_CREATOR,
    "Greater": Logical_Socket,
    "LayerNorm": DEFAULT_SOCKET_CREATOR,
    "LeakyRelu": DEFAULT_SOCKET_CREATOR,
    "Less": Logical_Socket,
    "LogSoftmax": DEFAULT_SOCKET_CREATOR,
    "MatMul": DEFAULT_SOCKET_CREATOR,
    "Max": DEFAULT_SOCKET_CREATOR,
    "MaxPool": DEFAULT_SOCKET_CREATOR,
    "Min": DEFAULT_SOCKET_CREATOR,
    "Mul": DEFAULT_SOCKET_CREATOR,
    "MultiHeadAttention": DEFAULT_SOCKET_CREATOR,
    "NonMaxSuppression": NonMaxSuppression_Socket,
    "NonZero": NonZero_Socket,
    "Not": DEFAULT_SOCKET_CREATOR,
    "Pad": Pad_Socket,
    "PRelu": DEFAULT_SOCKET_CREATOR,
    "Range": Range_Socket,
    "ReduceL2": DEFAULT_SOCKET_CREATOR,
    "ReduceMax": DEFAULT_SOCKET_CREATOR,
    "ReduceMean": DEFAULT_SOCKET_CREATOR,
    "ReduceSum": DEFAULT_SOCKET_CREATOR,
    "Relu": DEFAULT_SOCKET_CREATOR,
    "Reshape": Reshape_Socket,
    "Resize": Resize_Socket,
    "ScatterElements": ScatterElements_Socket,
    "ScatterND": ScatterND_Socket,
    "Shape": Shape_Socket,
    "Sigmoid": DEFAULT_SOCKET_CREATOR,
    "Slice": Slice_Socket,
    "Softmax": DEFAULT_SOCKET_CREATOR,
    "Softplus": DEFAULT_SOCKET_CREATOR,
    "Split": Split_Socket,
    "Squeeze": Squeeze_Socket,
    "Sub": DEFAULT_SOCKET_CREATOR,
    "Tile": Tile_Socket,
    "TopK": Topk_Socket,
    "Transpose": DEFAULT_SOCKET_CREATOR,
    "Unsqueeze": Unsqueeze_Socket,
    "Where": Where_Socket,
    "Sqrt": DEFAULT_SOCKET_CREATOR,
    "Log": DEFAULT_SOCKET_CREATOR,
    "Floor": DEFAULT_SOCKET_CREATOR,
    "RoiAlign": RoiAlign_Socket,
    "MMCVRoiAlign": RoiAlign_Socket,
    "SpaceToDepth": DEFAULT_SOCKET_CREATOR,
    "DepthToSpace": DEFAULT_SOCKET_CREATOR,
    "Scale": DEFAULT_SOCKET_CREATOR,  # caffe op
    "Tanh": DEFAULT_SOCKET_CREATOR,
    "Pow": DEFAULT_SOCKET_CREATOR,
    "Crop": DEFAULT_SOCKET_CREATOR,  # caffe op
    "ChannelShuffle": DEFAULT_SOCKET_CREATOR,  # caffe op
    "InstanceNormalization": DEFAULT_SOCKET_CREATOR,
    "Parameter": DEFAULT_SOCKET_CREATOR,  # caffe op
    "Interp": DEFAULT_SOCKET_CREATOR,  # caffe op
    "CaffeArgMax": DEFAULT_SOCKET_CREATOR,  # caffe op
    "HardSigmoid": DEFAULT_SOCKET_CREATOR,
    "HardSwish": DEFAULT_SOCKET_CREATOR,
    "Neg": DEFAULT_SOCKET_CREATOR,
    "GRU": DEFAULT_SOCKET_CREATOR,
    "PPQDeviceSwitch": DEFAULT_SOCKET_CREATOR,
    "Identity": DEFAULT_SOCKET_CREATOR,
    "OneHot": Onehot_Socket,
    "Reciprocal": DEFAULT_SOCKET_CREATOR,
    "GreaterOrEqual": Logical_Socket,
    "LessOrEqual": Logical_Socket,
    "Xor": Logical_Socket,
    "Or": Logical_Socket,
    "And": Logical_Socket,
    "Erf": DEFAULT_SOCKET_CREATOR,
}



import torch


def generate_indexer(
    num_of_fetches: int,
    num_of_elements: int,
    seed: int = 0x20211230) -> torch.Tensor:
    """Sample with a given seed. This function will generates a indexer based
    on your seed.

    Args:
        num_of_fetches (int): [description]
        num_of_elements (int): [description]
        seed (int, optional): [description]. Defaults to 0x20211230.

    Returns:
        torch.Tensor: [description]
    """

    indexer = []
    for i in range(num_of_fetches):
        indexer.append(seed % num_of_elements)
        seed = (0x343FD * seed + 0x269EC3) % (2 << 31)
    return torch.tensor(indexer, dtype=torch.int32)

def generate_torch_indexer(
    num_of_fetches: int,
    num_of_elements: int) -> torch.Tensor:
    return torch.randint(low=0, high=num_of_elements, size=[num_of_fetches])


def tensor_random_fetch(
    tensor: torch.Tensor, seed: int = None,
    num_of_fetches: int = 1024) -> torch.Tensor:
    """Fetch some elements from tensor randomly. if a valid seed is given,
    elements will be sampled based on your seed, otherwise a random seed will
    be generated.

    Args:
        tensor (torch.Tensor): [description]
        num_of_fetches (int, optional): [description]. Defaults to 1024.
    """
    tensor = tensor.flatten()
    num_of_elements = tensor.numel()
    assert num_of_elements > 0, ('Can not fetch data from empty tensor(0 element).')

    if seed is None:
        indexer = generate_torch_indexer(num_of_fetches=num_of_fetches, num_of_elements=num_of_elements)
    else: indexer = generate_indexer(num_of_fetches=num_of_fetches, num_of_elements=num_of_elements, seed=seed)
    return tensor.index_select(dim=0, index=indexer.to(tensor.device).long())


def channel_random_fetch(
    tensor: torch.Tensor,
    fetchs_per_channel: int = 1024,
    seed: int = None,
    channel_axis: int = 0) -> torch.Tensor:
    """Fetch some elements from tensor randomly by each channel. if a valid
    seed is given, elements will be sampled based on your seed, otherwise a
    random seed will be generated.

    result: [num_of_channel, fetchs_per_channel]

    Args:
        tensor (torch.Tensor): [description]
        fetchs_per_channel (int, optional): [description]. Defaults to 1024.
        channel_axis (int, optional): [description]. Defaults to 0.

    Returns:
        torch.Tensor: [description]
    """
    tensor = tensor.transpose(0, channel_axis)
    tensor = tensor.flatten(start_dim=1)
    num_of_elements = tensor.shape[-1]
    assert num_of_elements > 0, ('Can not fetch data from empty tensor(0 element).')

    if seed is None:
        indexer = generate_torch_indexer(num_of_fetches=fetchs_per_channel, num_of_elements=num_of_elements)
    else: indexer = generate_indexer(num_of_fetches=fetchs_per_channel, num_of_elements=num_of_elements, seed=seed)
    return tensor.index_select(dim=-1, index=indexer.to(tensor.device).long())


def batch_random_fetch(
    tensor: torch.Tensor,
    fetches_per_batch: int = 1024,
    seed: int = None
    ) -> torch.Tensor:
    """Fetch some elements from each sample in a batched tensor. if a valid
    seed is given, elements will be sampled based on your seed, otherwise a
    random seed will be generated.

    result: [num_of_batch, fetchs_per_batch]

    Args:
        tensor (torch.Tensor): [description]
        fetchs_per_channel (int, optional): [description]. Defaults to 1024.

    Returns:
        torch.Tensor: [description]
    """
    tensor = tensor.flatten(start_dim=1)
    num_of_elements = tensor.shape[-1]
    assert num_of_elements > 0, ('Can not fetch data from empty tensor(0 element).')

    if seed is None:
        indexer = generate_torch_indexer(num_of_fetches=fetches_per_batch, num_of_elements=num_of_elements)
    else: indexer = generate_indexer(num_of_fetches=fetches_per_batch, num_of_elements=num_of_elements, seed=seed)
    return tensor.index_select(dim=-1, index=indexer.to(tensor.device).long())



"""Regression tests for the local onnxslim wrapper."""

import importlib.util
import os
import sys
import types
import unittest
from unittest import mock

import onnx
from onnx import TensorProto, helper
from onnxconverter_common import float16 as convert_float_to_float16


def _load_onnxslim_pass_module():
    repo_root = os.path.join(os.path.dirname(__file__), "..", "xslim")

    package = types.ModuleType("xslim")
    package.__path__ = [repo_root]
    sys.modules["xslim"] = package

    logger_spec = importlib.util.spec_from_file_location(
        "xslim.logger", os.path.join(repo_root, "logger.py")
    )
    logger_module = importlib.util.module_from_spec(logger_spec)
    sys.modules["xslim.logger"] = logger_module
    logger_spec.loader.exec_module(logger_module)

    onnxslim_pass_spec = importlib.util.spec_from_file_location(
        "xslim.onnxslim_pass",
        os.path.join(repo_root, "onnxslim_pass", "__init__.py"),
        submodule_search_locations=[os.path.join(repo_root, "onnxslim_pass")],
    )
    onnxslim_pass_module = importlib.util.module_from_spec(onnxslim_pass_spec)
    sys.modules["xslim.onnxslim_pass"] = onnxslim_pass_module
    onnxslim_pass_spec.loader.exec_module(onnxslim_pass_module)
    return onnxslim_pass_module


def _load_onnx_graph_helper_module():
    repo_root = os.path.join(os.path.dirname(__file__), "..", "xslim")

    package = types.ModuleType("xslim")
    package.__path__ = [repo_root]
    sys.modules["xslim"] = package

    logger_spec = importlib.util.spec_from_file_location(
        "xslim.logger", os.path.join(repo_root, "logger.py")
    )
    logger_module = importlib.util.module_from_spec(logger_spec)
    sys.modules["xslim.logger"] = logger_module
    logger_spec.loader.exec_module(logger_module)

    defs_spec = importlib.util.spec_from_file_location(
        "xslim.defs", os.path.join(repo_root, "defs.py")
    )
    defs_module = importlib.util.module_from_spec(defs_spec)
    sys.modules["xslim.defs"] = defs_module
    defs_spec.loader.exec_module(defs_module)

    onnxslim_pass_module = _load_onnxslim_pass_module()
    sys.modules["xslim.onnxslim_pass"] = onnxslim_pass_module

    onnx_graph_helper_spec = importlib.util.spec_from_file_location(
        "xslim.onnx_graph_helper",
        os.path.join(repo_root, "onnx_graph_helper.py"),
    )
    onnx_graph_helper_module = importlib.util.module_from_spec(
        onnx_graph_helper_spec
    )
    sys.modules["xslim.onnx_graph_helper"] = onnx_graph_helper_module
    onnx_graph_helper_spec.loader.exec_module(onnx_graph_helper_module)
    return onnx_graph_helper_module


class TestOnnxSlimPass(unittest.TestCase):
    """Test compatibility behavior in the onnxslim wrapper."""

    @staticmethod
    def _build_add_model():
        x = helper.make_tensor_value_info("x", TensorProto.FLOAT, [1])
        y = helper.make_tensor_value_info("y", TensorProto.FLOAT, [1])
        z = helper.make_tensor_value_info("z", TensorProto.FLOAT, [1])
        bias = helper.make_tensor("bias", TensorProto.FLOAT, [1], [1.0])
        nodes = [
            helper.make_node("Add", ["x", "bias"], ["z"], name="add"),
            helper.make_node("Identity", ["z"], ["y"]),
        ]
        graph = helper.make_graph(
            nodes,
            "add_graph",
            [x],
            [y],
            [bias],
            value_info=[z],
        )
        model = helper.make_model(
            graph, opset_imports=[helper.make_opsetid("", 13)]
        )
        return onnx.shape_inference.infer_shapes(model)

    @staticmethod
    def _build_pad_pool_model(pool_op_type, pad_value=None):
        x = helper.make_tensor_value_info("x", TensorProto.FLOAT, [1, 1, 4, 4])
        y = helper.make_tensor_value_info("y", TensorProto.FLOAT, [1, 1, 4, 4])
        pads = helper.make_tensor(
            "pads", TensorProto.INT64, [8], [0, 0, 1, 1, 0, 0, 1, 1]
        )

        initializers = [pads]
        pad_inputs = ["x", "pads"]
        if pad_value is not None:
            pad_value_tensor = helper.make_tensor(
                "pad_value", TensorProto.FLOAT, [], [pad_value]
            )
            initializers.append(pad_value_tensor)
            pad_inputs.append("pad_value")

        nodes = [
            helper.make_node("Pad", pad_inputs, ["padded"], name="pad"),
            helper.make_node(
                pool_op_type,
                ["padded"],
                ["y"],
                name="pool",
                kernel_shape=[3, 3],
                strides=[1, 1],
            ),
        ]
        graph = helper.make_graph(
            nodes, "pad_pool_graph", [x], [y], initializers
        )
        model = helper.make_model(
            graph, opset_imports=[helper.make_opsetid("", 13)]
        )
        return onnx.shape_inference.infer_shapes(model)

    @staticmethod
    def _build_conv_model_without_kernel_shape(op_type):
        x = helper.make_tensor_value_info("x", TensorProto.FLOAT, [1, 3, 8, 8])
        y = helper.make_tensor_value_info("y", TensorProto.FLOAT, [1, 4, 6, 6])

        if op_type == "Conv":
            weight_shape = [4, 3, 3, 3]
        else:
            weight_shape = [3, 4, 3, 3]

        weight_count = (
            weight_shape[0]
            * weight_shape[1]
            * weight_shape[2]
            * weight_shape[3]
        )
        weight_values = [0.1] * int(weight_count)
        weight = helper.make_tensor(
            "w", TensorProto.FLOAT, weight_shape, weight_values
        )
        node = helper.make_node(
            op_type, ["x", "w"], ["y"], name=f"{op_type.lower()}_node"
        )
        graph = helper.make_graph(
            [node], f"{op_type.lower()}_graph", [x], [y], [weight]
        )
        return helper.make_model(
            graph, opset_imports=[helper.make_opsetid("", 13)]
        )

    def test_optimize_onnx_model_restores_shape_info_before_slim(self):
        onnxslim_pass_module = _load_onnxslim_pass_module()
        model = self._build_add_model()
        model = convert_float_to_float16.convert_float_to_float16(
            model, keep_io_types=True, disable_shape_infer=False
        )
        model.graph.ClearField("value_info")
        self.assertEqual(len(model.graph.value_info), 0)

        captured = {}

        def _capture_and_return(model_arg):
            captured["model"] = model_arg
            return model_arg

        with mock.patch.object(
            onnxslim_pass_module.onnxslim,
            "slim",
            side_effect=_capture_and_return,
        ):
            optimized_model = onnxslim_pass_module.optimize_onnx_model(model)
            optimized_model = onnxslim_pass_module.infer_onnx_model(
                optimized_model
            )

        self.assertIs(optimized_model, captured["model"])
        self.assertGreater(len(captured["model"].graph.value_info), 0)
        self.assertIn(
            TensorProto.FLOAT16,
            {
                value_info.type.tensor_type.elem_type
                for value_info in captured["model"].graph.value_info
            },
        )

    def test_optimize_onnx_model_fuses_pad_average_pool(self):
        onnxslim_pass_module = _load_onnxslim_pass_module()
        model = self._build_pad_pool_model("AveragePool")

        optimized_model = onnxslim_pass_module.optimize_onnx_model(model)
        optimized_model = onnxslim_pass_module.infer_onnx_model(
            optimized_model
        )

        onnx.checker.check_model(optimized_model)
        self.assertEqual(
            [node.op_type for node in optimized_model.graph.node],
            ["AveragePool"],
        )

        pool_node = optimized_model.graph.node[0]
        attrs = {
            attr.name: helper.get_attribute_value(attr)
            for attr in pool_node.attribute
        }
        self.assertEqual(list(attrs["pads"]), [1, 1, 1, 1])
        self.assertEqual(attrs["count_include_pad"], 1)

    def test_optimize_onnx_model_fuses_pad_max_pool_with_neg_inf_padding(
        self,
    ):
        onnxslim_pass_module = _load_onnxslim_pass_module()
        model = self._build_pad_pool_model("MaxPool", pad_value=float("-inf"))

        optimized_model = onnxslim_pass_module.optimize_onnx_model(model)
        optimized_model = onnxslim_pass_module.infer_onnx_model(
            optimized_model
        )

        onnx.checker.check_model(optimized_model)
        self.assertEqual(
            [node.op_type for node in optimized_model.graph.node], ["MaxPool"]
        )

        pool_node = optimized_model.graph.node[0]
        attrs = {
            attr.name: helper.get_attribute_value(attr)
            for attr in pool_node.attribute
        }
        self.assertEqual(list(attrs["pads"]), [1, 1, 1, 1])

    def test_optimize_onnx_model_keeps_zero_pad_before_max_pool(self):
        onnxslim_pass_module = _load_onnxslim_pass_module()
        model = self._build_pad_pool_model("MaxPool")

        optimized_model = onnxslim_pass_module.optimize_onnx_model(model)
        optimized_model = onnxslim_pass_module.infer_onnx_model(
            optimized_model
        )

        onnx.checker.check_model(optimized_model)
        self.assertEqual(
            [node.op_type for node in optimized_model.graph.node],
            ["Pad", "MaxPool"],
        )

    def test_format_onnx_model_fills_missing_conv_kernel_shape(self):
        onnx_graph_helper_module = _load_onnx_graph_helper_module()
        model = self._build_conv_model_without_kernel_shape("Conv")

        with mock.patch.object(
            onnx_graph_helper_module,
            "optimize_onnx_model",
            side_effect=lambda model_arg: model_arg,
        ), mock.patch.object(
            onnx_graph_helper_module,
            "infer_onnx_model",
            side_effect=lambda model_arg: model_arg,
        ):
            formatted_model = onnx_graph_helper_module.format_onnx_model(
                model, sim_en=False
            )

        conv_node = formatted_model.graph.node[0]
        attrs = {
            attr.name: helper.get_attribute_value(attr)
            for attr in conv_node.attribute
        }
        self.assertEqual(list(attrs["kernel_shape"]), [3, 3])

    def test_format_onnx_model_fills_missing_convtranspose_kernel_shape(self):
        onnx_graph_helper_module = _load_onnx_graph_helper_module()
        model = self._build_conv_model_without_kernel_shape("ConvTranspose")

        with mock.patch.object(
            onnx_graph_helper_module,
            "optimize_onnx_model",
            side_effect=lambda model_arg: model_arg,
        ), mock.patch.object(
            onnx_graph_helper_module,
            "infer_onnx_model",
            side_effect=lambda model_arg: model_arg,
        ):
            formatted_model = onnx_graph_helper_module.format_onnx_model(
                model, sim_en=False
            )

        conv_transpose_node = formatted_model.graph.node[0]
        attrs = {
            attr.name: helper.get_attribute_value(attr)
            for attr in conv_transpose_node.attribute
        }
        self.assertEqual(list(attrs["kernel_shape"]), [3, 3])


if __name__ == "__main__":
    unittest.main()



from .base.command import GraphCommand, GraphCommandType
from .base.graph import (BaseGraph, GraphBuilder, GraphExporter, Operation,
                         OperationExporter, Opset, Variable)
from .base.opdef import OperationBase, OpSocket, VLink
from .deploy import RunnableGraph
from .morph import GraphFormatter, GraphMerger, GraphReplacer
from .processer import DefaultGraphProcessor, GraphCommandProcessor
from .quantize import (DeviceSwitchOP, QuantableGraph, QuantableOperation,
                       QuantableVariable)
from .search import (Path, GraphPattern, SearchableGraph, TraversalCommand,
                     PatternMatchHelper)
from .training import TrainableGraph


#!/usr/bin/env python3
# Copyright (c) 2023 SpacemiT. All rights reserved.
import random
from collections import defaultdict
from enum import Enum
from typing import Callable, Dict, Iterable, List

import numpy as np
import torch
from tqdm import tqdm

from ..defs import XQUANT_CONFIG
from ..ppq_decorator import (
    BaseGraph,
    BaseGraphExecutor,
    LayerwiseEqualizationPass,
    Operation,
    QuantableOperation,
    TorchExecutor,
    empty_ppq_cache,
)

EQUALIZATION_OPERATION_TYPE = {"Conv", "Gemm", "ConvTranspose"}


class XSlimLayerwiseEqualizationPass(LayerwiseEqualizationPass):
    def __init__(
        self,
        iterations: int = XQUANT_CONFIG.equalization_iterations,
        value_threshold: float = 0.5,
        including_weight: bool = True,
        weight_multiplier: float = 1.0,
        including_bias: bool = True,
        including_act: bool = True,
        bias_multiplier: float = 0.5,
        act_multiplier: float = 0.5,
        interested_layers: List[str] = None,
        optimize_level: int = 2,
        verbose: bool = False,
    ) -> None:
        super().__init__(
            iterations,
            value_threshold,
            including_weight,
            weight_multiplier,
            including_bias,
            including_act,
            bias_multiplier,
            act_multiplier,
            interested_layers,
            optimize_level,
            verbose,
        )
        self.name = "XSlim LayerwiseEqualizationPass"

    def collect_activations(
        self,
        graph: BaseGraph,
        executor: TorchExecutor,
        dataloader: Iterable,
        collate_fn: Callable,
        operations: List[Operation],
        steps: int = 50,
    ) -> Dict[str, torch.Tensor]:
        def aggregate(op: Operation, tensor: torch.Tensor):
            if op.type in {"Conv", "ConvTranspose"}:  # Conv result: [n,c,h,w]
                num_of_channel = tensor.shape[1]
                tensor = tensor.transpose(0, 1)
                tensor = tensor.reshape(shape=[num_of_channel, -1])
                tensor = torch.max(tensor.abs(), dim=-1, keepdim=False)[0]
            elif op.type in {"MatMul", "Gemm"}:  # Gemm result: [n, c]
                tensor = tensor.transpose(0, 1)
                tensor = torch.max(tensor.abs(), dim=-1, keepdim=False)[0]
            return tensor

        output_names = []
        for operation in operations:
            assert operation.num_of_output == 1, f"Num of output of layer {operation.name} is supposed to be 1"
            output_names.append(operation.outputs[0].name)

        output_collector = defaultdict(list)
        executor.load_graph(graph)
        steps = min(steps, len(dataloader))
        loader_step_index = set(random.sample([i for i in range(len(dataloader))], steps))
        for idx, batch in enumerate(dataloader):
            if idx not in loader_step_index:
                continue
            if collate_fn is not None:
                data = collate_fn(batch)
            if isinstance(data, dict):
                data = {k: v.to(executor._executing_context.executing_device) for k, v in data.items()}
            elif isinstance(data, torch.Tensor):
                data = data.to(executor._executing_context.executing_device)

            outputs = executor.forward(data, output_names=output_names)
            for name, output in zip(output_names, outputs):
                op = graph.variables[name].source_op
                output_collector[name].append(aggregate(op, output).unsqueeze(-1))

        result = {}
        for name, output in zip(output_names, outputs):
            result[name] = torch.cat(output_collector[name], dim=-1)
        return result

    @empty_ppq_cache
    def optimize(
        self, graph: BaseGraph, dataloader: Iterable, executor: BaseGraphExecutor, collate_fn: Callable, **kwargs
    ) -> None:
        interested_operations = []

        if self.interested_layers is None:
            for operation in graph.operations.values():
                if operation.type in EQUALIZATION_OPERATION_TYPE:
                    interested_operations.append(operation)
        else:
            for name in self.interested_layers:
                if name in graph.operations:
                    interested_operations.append(graph.operations[name])

        pairs = self.find_equalization_pair(graph=graph, interested_operations=interested_operations)

        act_calib_steps = max(int(len(dataloader) / XQUANT_CONFIG.equalization_iterations), 50)

        for iter_times in tqdm(range(self.iterations), desc="Layerwise Equalization", total=self.iterations):
            if self.including_act:
                activations = self.collect_activations(
                    graph=graph,
                    executor=executor,
                    dataloader=dataloader,
                    collate_fn=collate_fn,
                    operations=interested_operations,
                    steps=act_calib_steps,
                )

                for name, act in activations.items():
                    graph.variables[name].value = act  # 将激活值写回网络

            for equalization_pair in pairs:
                equalization_pair.equalize(
                    value_threshold=self.value_threshold,
                    including_bias=self.including_bias,
                    including_act=self.including_act,
                    bias_multiplier=self.bias_multiplier,
                    act_multiplier=self.act_multiplier,
                )

        # equalization progress directly changes fp32 value of weight,
        # store it for following procedure.
        for op in graph.operations.values():
            if isinstance(op, QuantableOperation):
                op.store_parameter_value()
